from collections.abc import Callable, Mapping, MutableMapping, Sequence, Set
from datetime import UTC, date, datetime, time
from enum import Enum, IntEnum, StrEnum
from types import EllipsisType, NoneType, UnionType
from typing import Any, Literal, Self, Union, is_typeddict
from uuid import UUID

from haiway import MISSING, Missing
from haiway.state.attributes import (
    AttributeAnnotation,
    _resolve_type_typeddict,
)

from draive.commons import Meta
from draive.commons.metadata import validated_meta_value
from draive.parameters.types import (
    ParameterValidation,
    ParameterValidationContext,
    ParameterVerification,
)

__all__ = ("ParameterValidator",)


class ParameterValidator[Type]:
    @classmethod
    def of_typed_dict[Dict](
        cls,
        typed_dict: type[Dict],
        /,
    ) -> ParameterValidation[Dict]:
        if not is_typeddict(typed_dict):
            raise ValueError("Type has to be a TypedDict")

        return cls.of(
            _resolve_type_typeddict(
                typed_dict,
                module=typed_dict.__module__,
                type_parameters={},  # TODO: type parameters?
                self_annotation=None,
                recursion_guard={},
            ),
            verifier=None,
            recursion_guard={},
        )

    @classmethod
    def of(
        cls,
        annotation: AttributeAnnotation,
        /,
        *,
        verifier: ParameterVerification[Any] | None,
        recursion_guard: MutableMapping[str, ParameterValidation[Any]],
    ) -> ParameterValidation[Any]:
        if isinstance(annotation.origin, NotImplementedError | RuntimeError):
            raise annotation.origin  # raise an error if origin was not properly resolved

        if recursive := recursion_guard.get(str(annotation)):
            return recursive

        validator: Self = cls(
            annotation,
            validation=MISSING,
        )
        recursion_guard[str(annotation)] = validator

        if isinstance(annotation.origin, NotImplementedError | RuntimeError):
            raise annotation.origin  # raise an error if origin was not properly resolved

        if common := VALIDATORS.get(annotation.origin):
            validator.validation = common(annotation, verifier, recursion_guard)

        elif hasattr(annotation.origin, "model_validator"):
            validator.validation = annotation.origin.model_validator(verifier=verifier)

        elif is_typeddict(annotation.origin):
            validator.validation = _prepare_validator_of_typed_dict(
                annotation, verifier, recursion_guard
            )

        elif issubclass(annotation.origin, StrEnum):
            # TODO: FIXME: str enums !!!
            validator.validation = _prepare_validator_of_enum(annotation, verifier, recursion_guard)

        elif issubclass(annotation.origin, IntEnum):
            # TODO: FIXME: int enums !!!
            validator.validation = _prepare_validator_of_enum(annotation, verifier, recursion_guard)

        elif issubclass(annotation.origin, Enum):
            validator.validation = _prepare_validator_of_enum(annotation, verifier, recursion_guard)

        else:
            raise TypeError(f"Unsupported type annotation: {annotation}")

        return validator

    def __init__(
        self,
        annotation: AttributeAnnotation,
        validation: ParameterValidation[Type] | Missing,
    ) -> None:
        self.annotation: AttributeAnnotation = annotation
        self.validation: ParameterValidation[Type] | Missing = validation

    def __call__(
        self,
        value: Any,
        /,
        *,
        context: ParameterValidationContext,
    ) -> Any:
        assert self.validation is not MISSING  # nosec: B101
        return self.validation(  # pyright: ignore[reportCallIssue, reportUnknownVariableType]
            value,
            context=context,
        )


def _prepare_validator_of_any(
    annotation: AttributeAnnotation,
    /,
    verifier: ParameterVerification[Any] | None,
    recursion_guard: MutableMapping[str, ParameterValidation[Any]],
) -> ParameterValidation[Any]:
    def validator(
        value: Any,
        /,
        context: ParameterValidationContext,
    ) -> Any:
        return value  # any is always valid

    return validator


def _prepare_validator_of_none(
    annotation: AttributeAnnotation,
    /,
    verifier: ParameterVerification[Any] | None,
    recursion_guard: MutableMapping[str, ParameterValidation[Any]],
) -> ParameterValidation[Any]:
    if verifier := verifier:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            if value is None:
                verifier(value)
                return value

            else:
                raise TypeError(f"'{value}' is not matching expected type of 'None'")

    else:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            if value is None:
                return value

            else:
                raise TypeError(f"'{value}' is not matching expected type of 'None'")

    return validator


def _prepare_validator_of_missing(
    annotation: AttributeAnnotation,
    /,
    verifier: ParameterVerification[Any] | None,
    recursion_guard: MutableMapping[str, ParameterValidation[Any]],
) -> ParameterValidation[Any]:
    if verifier := verifier:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            if value is MISSING:
                verifier(value)
                return value

            else:
                raise TypeError(f"'{value}' is not matching expected type of 'Missing'")
    else:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            if value is MISSING:
                return value

            else:
                raise TypeError(f"'{value}' is not matching expected type of 'Missing'")

    return validator


def _prepare_validator_of_literal(
    annotation: AttributeAnnotation,
    /,
    verifier: ParameterVerification[Any] | None,
    recursion_guard: MutableMapping[str, ParameterValidation[Any]],
) -> ParameterValidation[Any]:
    elements: Sequence[Any] = annotation.arguments
    formatted_type: str = str(annotation)

    if verifier := verifier:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            if value in elements:
                verifier(value)
                return value

            else:
                raise ValueError(f"'{value}' is not matching expected values of '{formatted_type}'")

    else:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            if value in elements:
                return value

            else:
                raise ValueError(f"'{value}' is not matching expected values of '{formatted_type}'")

    return validator


def _prepare_validator_of_enum(
    annotation: AttributeAnnotation,
    /,
    verifier: ParameterVerification[Any] | None,
    recursion_guard: MutableMapping[str, ParameterValidation[Any]],
) -> ParameterValidation[Any]:
    elements: Sequence[Any] = list(annotation.origin)
    formatted_type: str = str(annotation)

    if verifier := verifier:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            if value in elements:
                verifier(value)
                return value

            else:
                raise ValueError(f"'{value}' is not matching expected values of '{formatted_type}'")

    else:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            if value in elements:
                return value

            else:
                raise ValueError(f"'{value}' is not matching expected values of '{formatted_type}'")

    return validator


def _prepare_validator_of_type(
    annotation: AttributeAnnotation,
    /,
    verifier: ParameterVerification[Any] | None,
    recursion_guard: MutableMapping[str, ParameterValidation[Any]],
) -> ParameterValidation[Any]:
    validated_type: type[Any] = annotation.origin
    formatted_type: str = str(annotation)

    if verifier := verifier:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            match value:
                case value if isinstance(value, validated_type):
                    verifier(value)
                    return value

                case _:
                    raise TypeError(
                        f"'{value}' is not matching expected type of '{formatted_type}'"
                    )
    else:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            match value:
                case value if isinstance(value, validated_type):
                    return value

                case _:
                    raise TypeError(
                        f"'{value}' is not matching expected type of '{formatted_type}'"
                    )

    return validator


def _prepare_validator_of_sequence(
    annotation: AttributeAnnotation,
    /,
    verifier: ParameterVerification[Any] | None,
    recursion_guard: MutableMapping[str, ParameterValidation[Any]],
) -> ParameterValidation[Any]:
    element_validator: ParameterValidation[Any] = ParameterValidator.of(
        annotation.arguments[0],
        verifier=None,
        recursion_guard=recursion_guard,
    )
    formatted_type: str = str(annotation)

    if verifier := verifier:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            match value:
                case [*elements]:
                    validated: Sequence[Any] = []
                    for idx, element in enumerate(elements):
                        with context.scope(f"[{idx}]"):
                            validated.append(element_validator(element, context=context))

                    validated = tuple(validated)
                    verifier(validated)
                    return validated

                case _:
                    raise TypeError(
                        f"'{value}' is not matching expected type of '{formatted_type}'"
                    )
    else:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            match value:
                case [*elements]:
                    validated: Sequence[Any] = []
                    for idx, element in enumerate(elements):
                        with context.scope(f"[{idx}]"):
                            validated.append(element_validator(element, context=context))

                    return tuple(validated)

                case _:
                    raise TypeError(
                        f"'{value}' is not matching expected type of '{formatted_type}'"
                    )

    return validator


def _prepare_validator_of_mapping(
    annotation: AttributeAnnotation,
    /,
    verifier: ParameterVerification[Any] | None,
    recursion_guard: MutableMapping[str, ParameterValidation[Any]],
) -> ParameterValidation[Any]:
    key_validator: ParameterValidation[Any] = ParameterValidator.of(
        annotation.arguments[0],
        verifier=None,
        recursion_guard=recursion_guard,
    )
    value_validator: ParameterValidation[Any] = ParameterValidator.of(
        annotation.arguments[1],
        verifier=None,
        recursion_guard=recursion_guard,
    )
    formatted_type: str = str(annotation)

    if verifier := verifier:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            match value:
                case {**elements}:
                    validated: MutableMapping[Any, Any] = {}
                    for key, element in elements.items():
                        with context.scope(f"[{key}]"):
                            validated[key_validator(key, context=context)] = value_validator(
                                element, context=context
                            )

                    # TODO: FIXME: make sure dict is not mutable?
                    # validated = MappingProxyType(validated)
                    verifier(validated)
                    return validated

                case _:
                    raise TypeError(
                        f"'{value}' is not matching expected type of '{formatted_type}'"
                    )
    else:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            match value:
                case {**elements}:
                    validated: MutableMapping[Any, Any] = {}
                    for key, element in elements.items():
                        with context.scope(f"[{key}]"):
                            validated[key_validator(key, context=context)] = value_validator(
                                element,
                                context=context,
                            )

                    # TODO: FIXME: make sure dict is not mutable?
                    # return MappingProxyType(validated)
                    return validated

                case _:
                    raise TypeError(
                        f"'{value}' is not matching expected type of '{formatted_type}'"
                    )

    return validator


def _prepare_validator_of_meta(  # noqa: C901
    annotation: AttributeAnnotation,
    /,
    verifier: ParameterVerification[Any] | None,
    recursion_guard: MutableMapping[str, ParameterValidation[Any]],
) -> ParameterValidation[Any]:
    formatted_type: str = str(annotation)

    if verifier := verifier:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            match value:
                case Meta() as meta:
                    return meta

                case {**elements}:
                    validated: MutableMapping[Any, Any] = {}
                    for key, element in elements.items():
                        if not isinstance(key, str):
                            raise TypeError(f"'{key}' is not matching expected type of '{str}'")

                        with context.scope(f"[{key}]"):
                            validated[key] = validated_meta_value(element)

                    meta: Meta = Meta(validated)
                    verifier(meta)
                    return meta

                case _:
                    raise TypeError(
                        f"'{value}' is not matching expected type of '{formatted_type}'"
                    )
    else:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            match value:
                case {**elements}:
                    validated: MutableMapping[Any, Any] = {}
                    for key, element in elements.items():
                        if not isinstance(key, str):
                            raise TypeError(f"'{key}' is not matching expected type of '{str}'")

                        with context.scope(f"[{key}]"):
                            validated[key] = validated_meta_value(element)

                    return Meta(validated)

                case _:
                    raise TypeError(
                        f"'{value}' is not matching expected type of '{formatted_type}'"
                    )

    return validator


def _prepare_validator_of_tuple(  # noqa: C901, PLR0915
    annotation: AttributeAnnotation,
    /,
    verifier: ParameterVerification[Any] | None,
    recursion_guard: MutableMapping[str, ParameterValidation[Any]],
) -> ParameterValidation[Any]:
    if (
        annotation.arguments[-1].origin == Ellipsis
        or annotation.arguments[-1].origin == EllipsisType
    ):
        element_validator: ParameterValidation[Any] = ParameterValidator.of(
            annotation.arguments[0],
            verifier=None,
            recursion_guard=recursion_guard,
        )
        formatted_type: str = str(annotation)

        if verifier := verifier:

            def validator(
                value: Any,
                /,
                context: ParameterValidationContext,
            ) -> Any:
                match value:
                    case [*elements]:
                        validated: Sequence[Any] = []
                        for idx, element in enumerate(elements):
                            with context.scope(f"[{idx}]"):
                                validated.append(element_validator(element, context=context))

                        validated = tuple(validated)
                        verifier(validated)
                        return validated

                    case _:
                        raise TypeError(
                            f"'{value}' is not matching expected type of '{formatted_type}'"
                        )
        else:

            def validator(
                value: Any,
                /,
                context: ParameterValidationContext,
            ) -> Any:
                match value:
                    case [*elements]:
                        validated: Sequence[Any] = []
                        for idx, element in enumerate(elements):
                            with context.scope(f"[{idx}]"):
                                validated.append(element_validator(element, context=context))

                        return tuple(validated)

                    case _:
                        raise TypeError(
                            f"'{value}' is not matching expected type of '{formatted_type}'"
                        )

        return validator

    else:
        element_validators: list[ParameterValidation[Any]] = [
            ParameterValidator.of(
                alternative,
                verifier=None,
                recursion_guard=recursion_guard,
            )
            for alternative in annotation.arguments
        ]
        elements_count: int = len(element_validators)
        formatted_type: str = str(annotation)

        if verifier := verifier:

            def validator(
                value: Any,
                /,
                context: ParameterValidationContext,
            ) -> Any:
                match value:
                    case [*elements]:
                        if len(elements) != elements_count:
                            raise ValueError(
                                f"'{value}' is not matching expected type of '{formatted_type}'"
                            )

                        validated: Sequence[Any] = []
                        for idx, element in enumerate(elements):
                            with context.scope(f"[{idx}]"):
                                validated.append(element_validators[idx](element, context=context))

                        validated = tuple(validated)
                        verifier(validated)
                        return validated

                    case _:
                        raise TypeError(
                            f"'{value}' is not matching expected type of '{formatted_type}'"
                        )
        else:

            def validator(
                value: Any,
                /,
                context: ParameterValidationContext,
            ) -> Any:
                match value:
                    case [*elements]:
                        if len(elements) != elements_count:
                            raise ValueError(
                                f"'{value}' is not matching expected type of '{formatted_type}'"
                            )

                        validated: Sequence[Any] = []
                        for idx, element in enumerate(elements):
                            with context.scope(f"[{idx}]"):
                                validated.append(element_validators[idx](element, context=context))

                        return tuple(validated)

                    case _:
                        raise TypeError(
                            f"'{value}' is not matching expected type of '{formatted_type}'"
                        )

        return validator


def _prepare_validator_of_union(
    annotation: AttributeAnnotation,
    /,
    verifier: ParameterVerification[Any] | None,
    recursion_guard: MutableMapping[str, ParameterValidation[Any]],
) -> ParameterValidation[Any]:
    validators: list[ParameterValidation[Any]] = [
        ParameterValidator.of(
            alternative,
            verifier=None,
            recursion_guard=recursion_guard,
        )
        for alternative in annotation.arguments
    ]
    formatted_type: str = str(annotation)

    if verifier := verifier:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            errors: list[Exception] = []
            for validator in validators:
                try:
                    validated = validator(value, context=context)
                    verifier(validated)
                    return validated

                except Exception as exc:
                    errors.append(exc)

            raise ExceptionGroup(
                f"'{value}' is not matching expected type of '{formatted_type}'",
                errors,
            )
    else:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            errors: list[Exception] = []
            for validator in validators:
                try:
                    return validator(value, context=context)

                except Exception as exc:
                    errors.append(exc)

            raise ExceptionGroup(
                f"'{value}' is not matching expected type of '{formatted_type}'",
                errors,
            )

    return validator


def _prepare_validator_of_float(  # noqa: C901
    annotation: AttributeAnnotation,
    /,
    verifier: ParameterVerification[Any] | None,
    recursion_guard: MutableMapping[str, ParameterValidation[Any]],
) -> ParameterValidation[Any]:
    if verifier := verifier:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            match value:
                case float():
                    verifier(value)
                    return value

                case int():
                    validated = float(value)
                    verifier(validated)
                    return validated

                case str():
                    try:
                        validated = float(value)

                    except (ValueError, TypeError):
                        pass  # invalid

                    else:
                        verifier(validated)
                        return validated

                case _:
                    pass  # invalid

            raise TypeError(f"'{value}' is not matching expected type of 'float'")

    else:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            match value:
                case float():
                    return value

                case int():
                    return float(value)

                case str():
                    try:
                        return float(value)

                    except (ValueError, TypeError):
                        pass  # invalid

                case _:
                    pass  # invalid

            raise TypeError(f"'{value}' is not matching expected type of 'float'")

    return validator


def _prepare_validator_of_int(  # noqa: C901
    annotation: AttributeAnnotation,
    /,
    verifier: ParameterVerification[Any] | None,
    recursion_guard: MutableMapping[str, ParameterValidation[Any]],
) -> ParameterValidation[Any]:
    if verifier := verifier:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            match value:
                case int():
                    verifier(value)
                    return value

                case float():
                    validated = int(value)
                    if value == validated:
                        verifier(validated)
                        return int(validated)

                    # else invalid

                case str():
                    try:
                        validated = int(value)

                    except (ValueError, TypeError):
                        pass  # invalid

                    else:
                        verifier(validated)
                        return validated

                case _:
                    pass  # invalid

            raise TypeError(f"'{value}' is not matching expected type of 'int'")

    else:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            match value:
                case int():
                    return value

                case float():
                    validated = int(value)
                    if value == validated:
                        return int(validated)

                    # else invalid

                case str():
                    try:
                        return int(value)

                    except (ValueError, TypeError):
                        pass  # invalid

                case _:
                    pass  # invalid

            raise TypeError(f"'{value}' is not matching expected type of 'int'")

    return validator


def _prepare_validator_of_bool(  # noqa: C901
    annotation: AttributeAnnotation,
    /,
    verifier: ParameterVerification[Any] | None,
    recursion_guard: MutableMapping[str, ParameterValidation[Any]],
) -> ParameterValidation[Any]:
    if verifier := verifier:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            match value:
                case bool():
                    verifier(value)
                    return value

                case int():
                    validated = bool(value != 0)
                    verifier(validated)
                    return validated

                case str():
                    match value.lower():
                        case "true" | "t" | "1":
                            verifier(True)
                            return True

                        case "false" | "f" | "0":
                            verifier(False)
                            return False

                        case _:
                            pass  # invalid

                case _:
                    pass  # invalid

            raise TypeError(f"'{value}' is not matching expected type of 'bool'")
    else:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            match value:
                case bool():
                    return value

                case int():
                    validated = bool(value != 0)
                    return validated

                case str():
                    match value.lower():
                        case "true" | "t" | "1":
                            return True

                        case "false" | "f" | "0":
                            return False

                        case _:
                            pass  # invalid

                case _:
                    pass  # invalid

            raise TypeError(f"'{value}' is not matching expected type of 'bool'")

    return validator


def _prepare_validator_of_uuid(
    annotation: AttributeAnnotation,
    /,
    verifier: ParameterVerification[Any] | None,
    recursion_guard: MutableMapping[str, ParameterValidation[Any]],
) -> ParameterValidation[Any]:
    if verifier := verifier:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            if isinstance(value, UUID):
                verifier(value)
                return value

            elif isinstance(value, str):
                validated = UUID(hex=value)
                verifier(validated)
                return validated

            raise TypeError(f"'{value}' is not matching expected type of 'UUID'")
    else:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            if isinstance(value, UUID):
                return value

            elif isinstance(value, str):
                return UUID(hex=value)

            raise TypeError(f"'{value}' is not matching expected type of 'UUID'")

    return validator


def _prepare_validator_of_datetime(
    annotation: AttributeAnnotation,
    /,
    verifier: ParameterVerification[Any] | None,
    recursion_guard: MutableMapping[str, ParameterValidation[Any]],
) -> ParameterValidation[Any]:
    if verifier := verifier:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            if isinstance(value, datetime):
                verifier(value)
                return value

            elif isinstance(value, str):
                validated = datetime.fromisoformat(value)
                verifier(validated)
                return validated

            elif isinstance(value, int | float):
                validated = datetime.fromtimestamp(
                    value,
                    UTC,
                )
                verifier(validated)
                return validated

            raise TypeError(f"'{value}' is not matching expected type of 'datetime'")
    else:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            if isinstance(value, datetime):
                return value

            elif isinstance(value, str):
                return datetime.fromisoformat(value)

            elif isinstance(value, int | float):
                return datetime.fromtimestamp(
                    value,
                    UTC,
                )

            raise TypeError(f"'{value}' is not matching expected type of 'datetime'")

    return validator


def _prepare_validator_of_date(
    annotation: AttributeAnnotation,
    /,
    verifier: ParameterVerification[Any] | None,
    recursion_guard: MutableMapping[str, ParameterValidation[Any]],
) -> ParameterValidation[Any]:
    if verifier := verifier:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            if isinstance(value, date):
                verifier(value)
                return value

            elif isinstance(value, str):
                validated = date.fromisoformat(value)
                verifier(validated)
                return validated

            elif isinstance(value, int | float):
                validated = date.fromtimestamp(value)
                verifier(validated)
                return validated

            raise TypeError(f"'{value}' is not matching expected type of 'datetime'")
    else:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            if isinstance(value, date):
                return value

            elif isinstance(value, str):
                return date.fromisoformat(value)

            elif isinstance(value, int | float):
                return date.fromtimestamp(value)

            raise TypeError(f"'{value}' is not matching expected type of 'datetime'")

    return validator


def _prepare_validator_of_time(
    annotation: AttributeAnnotation,
    /,
    verifier: ParameterVerification[Any] | None,
    recursion_guard: MutableMapping[str, ParameterValidation[Any]],
) -> ParameterValidation[Any]:
    if verifier := verifier:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            if isinstance(value, datetime):
                verifier(value)
                return value

            elif isinstance(value, str):
                validated = datetime.fromisoformat(value)
                verifier(validated)
                return validated

            elif isinstance(value, int | float):
                validated = datetime.fromtimestamp(
                    value,
                    UTC,
                )
                verifier(validated)
                return validated

            raise TypeError(f"'{value}' is not matching expected type of 'datetime'")
    else:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            if isinstance(value, datetime):
                return value

            elif isinstance(value, str):
                return datetime.fromisoformat(value)

            elif isinstance(value, int | float):
                return datetime.fromtimestamp(
                    value,
                    UTC,
                )

            raise TypeError(f"'{value}' is not matching expected type of 'datetime'")

    return validator


def _prepare_validator_of_typed_dict(  # noqa: C901
    annotation: AttributeAnnotation,
    /,
    verifier: ParameterVerification[Any] | None,
    recursion_guard: MutableMapping[str, ParameterValidation[Any]],
) -> ParameterValidation[Any]:
    def key_validator(
        value: Any,
    ) -> str:
        match value:
            case value if isinstance(value, str):
                return value

            case _:
                raise TypeError(f"'{value}' is not matching expected type of 'str'")

    formatted_type: str = str(annotation)
    values_validators: dict[str, ParameterValidation[Any]] = {
        key: ParameterValidator.of(
            element,
            verifier=None,
            recursion_guard=recursion_guard,
        )
        for key, element in annotation.extra["attributes"].items()
    }
    required_values: Set[str] = annotation.extra["required"]

    if verifier := verifier:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            match value:
                case {**elements}:
                    validated: MutableMapping[Any, Any] = {}
                    for key, validate in values_validators.items():
                        validated_key: str = key_validator(key)
                        with context.scope(f"[{validated_key}]"):
                            element: Any = elements.get(validated_key, MISSING)
                            if element is MISSING and key not in required_values:
                                continue  # skip missing and not required

                            validated[validated_key] = validate(
                                element,
                                context=context,
                            )

                    # TODO: FIXME: make sure dict is not mutable?
                    # validated = MappingProxyType(validated)
                    verifier(validated)
                    return validated

                case _:
                    raise TypeError(
                        f"'{value}' is not matching expected type of '{formatted_type}'"
                    )
    else:

        def validator(
            value: Any,
            /,
            context: ParameterValidationContext,
        ) -> Any:
            match value:
                case {**elements}:
                    validated: MutableMapping[Any, Any] = {}
                    for key, validate in values_validators.items():
                        validated_key: str = key_validator(key)
                        with context.scope(f"[{validated_key}]"):
                            element: Any = elements.get(validated_key, MISSING)
                            if element is MISSING and key not in required_values:
                                continue  # skip missing and not required

                            validated[validated_key] = validate(
                                element,
                                context=context,
                            )

                    # TODO: FIXME: make sure dict is not mutable?
                    # return MappingProxyType(validated)
                    return validated

                case _:
                    raise TypeError(
                        f"'{value}' is not matching expected type of '{formatted_type}'"
                    )

    return validator


VALIDATORS: Mapping[
    Any,
    Callable[
        [
            AttributeAnnotation,
            ParameterVerification[Any] | None,
            MutableMapping[str, ParameterValidation[Any]],
        ],
        ParameterValidation[Any],
    ],
] = {
    Any: _prepare_validator_of_any,
    NoneType: _prepare_validator_of_none,
    Missing: _prepare_validator_of_missing,
    type: _prepare_validator_of_type,
    bool: _prepare_validator_of_bool,
    int: _prepare_validator_of_int,
    float: _prepare_validator_of_float,
    str: _prepare_validator_of_type,
    tuple: _prepare_validator_of_tuple,
    Literal: _prepare_validator_of_literal,
    Sequence: _prepare_validator_of_sequence,
    Mapping: _prepare_validator_of_mapping,
    Meta: _prepare_validator_of_meta,
    UUID: _prepare_validator_of_uuid,
    date: _prepare_validator_of_date,
    datetime: _prepare_validator_of_datetime,
    time: _prepare_validator_of_time,
    Union: _prepare_validator_of_union,
    UnionType: _prepare_validator_of_union,
}



from typing import cast

from draive.evaluation import EvaluationScore, EvaluationScoreValue, evaluator
from draive.multimodal import Multimodal, MultimodalContent, MultimodalTagElement
from draive.stages import Stage

__all__ = ("truthfulness_evaluator",)


INSTRUCTION: str = """\
You are evaluating the provided content according to the defined criteria.

<INSTRUCTION>
Compare the REFERENCE and the EVALUATED content by carefully examining them, then rate\
 the EVALUATED content using solely a truthfulness (factual accuracy) metric according to the EVALUATION_CRITERIA.
Think step by step and provide explanation of the score before the final score.
Use the explained RATING scale and the requested FORMAT to provide the result.
</INSTRUCTION>

<EVALUATION_CRITERIA>
Evaluated metric is truthfulness (factual accuracy) - this metric evaluates the factual correctness\
 of the content, regardless of its relation to the source material. Truthfulness is concerned with whether\
 the information presented is accurate in the broader context, ensuring that the facts are correct and reliable,\
 and that the content does not perpetuate errors or falsehoods. This metric is more concerned with the overall\
 factual integrity of the content, meaning that even if a claim is not directly traceable to the provided source,\
 it must still be correct and represent the truth of the matter. Truthfulness is about the correctness of the\
 facts themselves, not just their alignment with a specific source.
</EVALUATION_CRITERIA>
{guidelines}
<RATING>
Assign a truthfulness (factual accuracy) score using exact name of one of the following values:
- "poor" is very low truthfulness, the content contains many factual inaccuracies or distortions.
- "fair" is low truthfulness, the content has some accurate information but also significant factual errors or misrepresentations.
- "good" is moderate truthfulness, the content is somewhat accurate but with noticeable factual inaccuracies or distortions.
- "excellent" is high truthfulness, the content is mostly accurate with minimal factual errors or misrepresentations.
- "perfect" is very high truthfulness, the content is fully accurate, with no factual errors or misrepresentations.
Use the "none" value for content that cannot be rated at all.
</RATING>

<FORMAT>
The final result containing only the numerical score value HAVE to be put inside a `RESULT`\
 xml tag within the result i.e. `<RESULT>good</RESULT>`.
</FORMAT>
"""  # noqa: E501


@evaluator(name="truthfulness")
async def truthfulness_evaluator(
    evaluated: Multimodal,
    /,
    *,
    reference: Multimodal,
    guidelines: str | None = None,
) -> EvaluationScore:
    if not evaluated:
        return EvaluationScore(
            value=0,
            comment="Input was empty!",
        )

    if not reference:
        return EvaluationScore(
            value=0,
            comment="Reference was empty!",
        )

    completion: MultimodalContent = await Stage.completion(
        MultimodalContent.of(
            "<REFERENCE>",
            reference,
            "</REFERENCE>\n<EVALUATED>",
            evaluated,
            "</EVALUATED>",
        ),
        instruction=INSTRUCTION.format(
            guidelines=f"\n<GUIDELINES>\n{guidelines}\n</GUIDELINES>\n"
            if guidelines is not None
            else ""
        ),
    ).execute()

    if result := MultimodalTagElement.parse_first(
        "RESULT",
        content=completion,
    ):
        return EvaluationScore.of(
            cast(EvaluationScoreValue, result.content.to_str()),
            comment=completion.to_str(),
        )

    else:
        raise ValueError("Invalid evaluator result:\n%s", completion)



from collections.abc import AsyncIterator, Iterable, Sequence
from typing import Any, Literal, cast, overload

from anthropic import NOT_GIVEN
from anthropic import RateLimitError as AnthropicRateLimitError
from anthropic.types import (
    Message,
    MessageParam,
    RedactedThinkingBlock,
    TextBlock,
    ThinkingBlock,
    ToolUseBlock,
)
from haiway import MISSING, ObservabilityLevel, as_list, ctx

from draive.anthropic.api import AnthropicAPI
from draive.anthropic.config import AnthropicConfig
from draive.anthropic.lmm import (
    content_block_as_content_element,
    context_element_as_message,
    convert_content_element,
    output_as_response_declaration,
    thinking_budget_as_config,
    tools_as_tool_config,
)
from draive.anthropic.types import AnthropicException
from draive.anthropic.utils import unwrap_missing
from draive.commons import META_EMPTY
from draive.instructions import Instruction
from draive.lmm import (
    LMMCompletion,
    LMMContext,
    LMMOutput,
    LMMOutputSelection,
    LMMStreamOutput,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolSelection,
    LMMToolSpecification,
)
from draive.lmm.state import LMM
from draive.multimodal import Multimodal, MultimodalContent
from draive.utils import RateLimitError

__all__ = ("AnthropicLMMGeneration",)


class AnthropicLMMGeneration(AnthropicAPI):
    def lmm(self) -> LMM:
        return LMM(completing=self.lmm_completion)

    @overload
    async def lmm_completion(
        self,
        *,
        instruction: Instruction | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        prefill: Multimodal | None = None,
        config: AnthropicConfig | None = None,
        output: LMMOutputSelection,
        stream: Literal[False] = False,
        **extra: Any,
    ) -> LMMOutput: ...

    @overload
    async def lmm_completion(
        self,
        *,
        instruction: Instruction | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        output: LMMOutputSelection,
        prefill: Multimodal | None = None,
        config: AnthropicConfig | None = None,
        stream: Literal[True],
        **extra: Any,
    ) -> AsyncIterator[LMMStreamOutput]: ...

    async def lmm_completion(  # noqa: C901, PLR0912, PLR0915
        self,
        *,
        instruction: Instruction | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        output: LMMOutputSelection,
        prefill: Multimodal | None = None,
        config: AnthropicConfig | None = None,
        stream: bool = False,
        **extra: Any,
    ) -> AsyncIterator[LMMStreamOutput] | LMMOutput:
        if stream:
            raise NotImplementedError("Anthropic streaming is not implemented yet")

        completion_config: AnthropicConfig = config or ctx.state(AnthropicConfig)
        with ctx.scope("anthropic_lmm_completion", completion_config):
            ctx.record(
                ObservabilityLevel.INFO,
                attributes={
                    "lmm.provider": "anthropic",
                    "lmm.model": completion_config.model,
                    "lmm.temperature": completion_config.temperature,
                    "lmm.max_tokens": completion_config.max_tokens,
                    "lmm.tools": [tool["name"] for tool in tools] if tools else None,
                    "lmm.tool_selection": f"{tool_selection}" if tools else None,
                    "lmm.stream": stream,
                    "lmm.output": f"{output}",
                    "lmm.instruction": f"{instruction}",
                    "lmm.context": [element.to_str() for element in context],
                },
            )

            messages: list[MessageParam] = [
                context_element_as_message(element) for element in context
            ]

            response_prefill, output_decoder = output_as_response_declaration(
                output=output,
                prefill=prefill,
            )

            if response_prefill:  # use custom prefill
                messages.append(
                    {
                        "role": "assistant",
                        "content": [
                            convert_content_element(element) for element in response_prefill.parts
                        ],
                    }
                )

            tool_choice, tools_list = tools_as_tool_config(
                tools,
                tool_selection=tool_selection,
            )

            completion: Message
            try:
                completion = await self._client.messages.create(
                    model=completion_config.model,
                    system=Instruction.formatted(instruction) if instruction else NOT_GIVEN,
                    messages=messages,
                    temperature=completion_config.temperature,
                    top_p=unwrap_missing(completion_config.top_p),
                    max_tokens=completion_config.max_tokens,
                    thinking=thinking_budget_as_config(completion_config.thinking_tokens_budget)
                    if self._provider == "anthropic"
                    else NOT_GIVEN,  # bedrock does not support thinking yet
                    tools=tools_list,
                    tool_choice=tool_choice,
                    stop_sequences=as_list(cast(Sequence[str], completion_config.stop_sequences))
                    if completion_config.stop_sequences is not MISSING
                    else NOT_GIVEN,
                    stream=False,
                )

            except AnthropicRateLimitError as exc:  # retry on rate limit after delay
                if delay := exc.response.headers.get("Retry-After"):
                    ctx.record(
                        ObservabilityLevel.WARNING,
                        event="lmm.rate_limit",
                        attributes={"delay": delay},
                    )
                    try:
                        raise RateLimitError(retry_after=float(delay)) from exc

                    except ValueError:
                        raise exc from None

                else:
                    ctx.record(
                        ObservabilityLevel.WARNING,
                        event="lmm.rate_limit",
                    )
                    raise exc

            ctx.record(
                ObservabilityLevel.INFO,
                metric="lmm.input_tokens",
                value=completion.usage.input_tokens,
                unit="tokens",
                attributes={"lmm.model": completion.model},
            )
            ctx.record(
                ObservabilityLevel.INFO,
                metric="lmm.input_tokens.cached",
                value=completion.usage.cache_creation_input_tokens or 0,
                unit="tokens",
                attributes={"lmm.model": completion.model},
            )
            ctx.record(
                ObservabilityLevel.INFO,
                metric="lmm.output_tokens",
                value=completion.usage.output_tokens,
                unit="tokens",
                attributes={"lmm.model": completion.model},
            )
            ctx.record(
                ObservabilityLevel.INFO,
                metric="lmm.output_tokens.cached",
                value=completion.usage.cache_read_input_tokens or 0,
                unit="tokens",
                attributes={"lmm.model": completion.model},
            )

            match completion.stop_reason:
                case "end_turn" | "tool_use" | "stop_sequence":
                    pass  # process results

                case "max_tokens":
                    raise AnthropicException("Invalid Anthropic completion - tokens limit!")

                case reason:
                    raise AnthropicException(f"Anthropic completion generation failed - {reason}!")

            message_parts: list[TextBlock | ThinkingBlock | RedactedThinkingBlock] = []
            tool_calls: list[ToolUseBlock] = []
            for part in completion.content:
                match part:
                    case TextBlock() as text:
                        message_parts.append(text)

                    case ToolUseBlock() as call:
                        tool_calls.append(call)

                    case ThinkingBlock() as thinking:
                        message_parts.append(thinking)

                    case RedactedThinkingBlock() as redacted_thinking:
                        message_parts.append(redacted_thinking)

                    case _:
                        pass  # skip unknown elements

            lmm_completion: LMMCompletion | None
            if message_parts:
                if response_prefill:
                    lmm_completion = LMMCompletion.of(
                        output_decoder(
                            MultimodalContent.of(
                                response_prefill,
                                *[content_block_as_content_element(part) for part in message_parts],
                            )
                        ),
                    )

                else:
                    lmm_completion = LMMCompletion.of(
                        output_decoder(
                            MultimodalContent.of(
                                *[content_block_as_content_element(part) for part in message_parts],
                            )
                        ),
                    )

            else:
                lmm_completion = None

            if tool_calls:
                assert tools, "Requesting tool call without tools"  # nosec: B101
                completion_tool_calls = LMMToolRequests(
                    content=lmm_completion.content if lmm_completion else None,
                    requests=[
                        LMMToolRequest(
                            identifier=call.id,
                            tool=call.name,
                            arguments=cast(dict[str, Any], call.input),
                        )
                        for call in tool_calls
                    ],
                    meta=META_EMPTY,
                )
                ctx.record(
                    ObservabilityLevel.INFO,
                    event="lmm.tool_requests",
                    attributes={"lmm.tools": [call.name for call in tool_calls]},
                )
                return completion_tool_calls

            elif lmm_completion:
                ctx.record(
                    ObservabilityLevel.INFO,
                    event="lmm.completion",
                )
                return lmm_completion

            else:
                raise AnthropicException(
                    "Invalid Anthropic completion, missing content!", completion
                )



from collections.abc import Sequence
from typing import Literal

from haiway import MISSING, Missing

from draive.configuration import Config

__all__ = (
    "VLLMChatConfig",
    "VLLMEmbeddingConfig",
)


class VLLMChatConfig(Config):
    model: str
    temperature: float = 1.0
    top_p: float | Missing = MISSING
    frequency_penalty: float | Missing = MISSING
    max_tokens: int | Missing = MISSING
    seed: int | Missing = MISSING
    vision_details: Literal["auto", "low", "high"] | Missing = MISSING
    parallel_tool_calls: bool | Missing = MISSING
    timeout: float | Missing = MISSING
    stop_sequences: Sequence[str] | Missing = MISSING


class VLLMEmbeddingConfig(Config):
    model: str
    dimensions: int | Missing = MISSING
    batch_size: int = 128
    timeout: float | Missing = MISSING



from draive.guardrails.quality.state import GuardrailsQualityVerification
from draive.guardrails.quality.types import (
    GuardrailsQualityException,
    GuardrailsQualityVerifying,
)

__all__ = (
    "GuardrailsQualityException",
    "GuardrailsQualityVerification",
    "GuardrailsQualityVerifying",
)



from collections import deque
from collections.abc import AsyncIterator

__all__ = (
    "ConstantStream",
    "FixedStream",
)


class ConstantStream[Value](AsyncIterator[Value]):
    def __init__(
        self,
        value: Value,
        /,
    ) -> None:
        self._value: Value = value

    async def __anext__(self) -> Value:
        return self._value


class FixedStream[Value](AsyncIterator[Value]):
    def __init__(
        self,
        *values: Value,
    ) -> None:
        self._values: deque[Value] = deque(values)

    async def __anext__(self) -> Value:
        if self._values:
            return self._values.popleft()

        else:
            raise StopAsyncIteration()



from collections.abc import Sequence
from typing import Any, Literal, overload

from haiway import State, ctx

from draive.commons import META_EMPTY, Meta
from draive.resources.types import (
    Resource,
    ResourceDeclaration,
    ResourceFetching,
    ResourceListFetching,
    ResourceMissing,
    ResourceUploading,
)

__all__ = ("Resources",)


class Resources(State):
    @classmethod
    async def fetch_list(
        cls,
        **extra: Any,
    ) -> Sequence[ResourceDeclaration]:
        return await ctx.state(cls).list_fetching(**extra)

    @overload
    @classmethod
    async def fetch(
        cls,
        resource: ResourceDeclaration | str,
        /,
    ) -> Resource | None: ...

    @overload
    @classmethod
    async def fetch(
        cls,
        resource: ResourceDeclaration | str,
        /,
        *,
        default: Resource,
    ) -> Resource: ...

    @overload
    @classmethod
    async def fetch(
        cls,
        resource: ResourceDeclaration | str,
        /,
        *,
        required: Literal[True],
    ) -> Resource: ...

    @classmethod
    async def fetch(
        cls,
        resource: ResourceDeclaration | str,
        /,
        *,
        default: Resource | None = None,
        required: bool = True,
    ) -> Resource | None:
        uri: str = resource if isinstance(resource, str) else resource.uri

        if fetched := await ctx.state(cls).fetching(uri):
            return fetched

        elif required and default is None:
            raise ResourceMissing(f"Missing resource: '{uri}'")

        else:
            return default

    @classmethod
    async def upload(
        cls,
        resource: Resource,
        /,
        **extra: Any,
    ) -> None:
        return await ctx.state(cls).uploading(resource, **extra)

    list_fetching: ResourceListFetching
    fetching: ResourceFetching
    uploading: ResourceUploading
    meta: Meta = META_EMPTY



from collections import deque
from collections.abc import Callable, Mapping, Sequence
from types import TracebackType
from typing import Any, Protocol

__all__ = (
    "BasicValue",
    "ParameterConversion",
    "ParameterValidation",
    "ParameterValidationContext",
    "ParameterValidationError",
    "ParameterVerification",
)

type BasicValue = (
    Mapping[str, "BasicValue"] | Sequence["BasicValue"] | str | float | int | bool | None
)

type ParameterConversion[Type] = Callable[[Type], BasicValue]
type ParameterVerification[Type] = Callable[[Type], None]


class ParameterValidationContext:
    def __init__(self) -> None:
        self._path: deque[str] = deque()

    def __str__(self) -> str:
        return "".join(self._path)

    def scope(
        self,
        path: str,
        /,
    ) -> "ParameterValidationContextScope":
        return ParameterValidationContextScope(self, component=path)


class ParameterValidationContextScope:
    def __init__(
        self,
        context: ParameterValidationContext,
        /,
        component: str,
    ) -> None:
        self.context: ParameterValidationContext = context
        self.component: str = component

    def __enter__(self) -> ParameterValidationContext:
        self.context._path.append(self.component)
        return self.context

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        try:
            if exc_val is not None and exc_type is not ParameterValidationError:
                raise ParameterValidationError(
                    f"Validation error at {self.context!s}",
                ) from exc_val

        finally:
            self.context._path.pop()


class ParameterValidation[Type](Protocol):
    def __call__(
        self,
        value: Any,
        /,
        *,
        context: ParameterValidationContext,
    ) -> Type: ...


class ParameterValidationError(Exception):
    pass



from collections.abc import AsyncIterator
from typing import Any, Literal, Protocol, overload, runtime_checkable

from draive.conversation.types import (
    ConversationMemory,
    ConversationMessage,
    ConversationStreamElement,
)
from draive.instructions import Instruction
from draive.multimodal import (
    Multimodal,
)
from draive.tools import Toolbox

__all__ = ("ConversationCompleting",)


@runtime_checkable
class ConversationCompleting(Protocol):
    @overload
    async def __call__(
        self,
        *,
        instruction: Instruction | None,
        input: ConversationMessage | Multimodal,
        memory: ConversationMemory,
        toolbox: Toolbox,
        stream: Literal[False] = False,
        **extra: Any,
    ) -> ConversationMessage: ...

    @overload
    async def __call__(
        self,
        *,
        instruction: Instruction | None,
        input: ConversationMessage | Multimodal,
        memory: ConversationMemory,
        toolbox: Toolbox,
        stream: Literal[True],
        **extra: Any,
    ) -> AsyncIterator[ConversationStreamElement]: ...

    async def __call__(
        self,
        *,
        instruction: Instruction | None,
        input: ConversationMessage | Multimodal,  # noqa: A002
        memory: ConversationMemory,
        toolbox: Toolbox,
        stream: bool = False,
        **extra: Any,
    ) -> AsyncIterator[ConversationStreamElement] | ConversationMessage: ...



from collections.abc import Mapping, Sequence
from typing import Any, final

from draive.instructions.types import (
    Instruction,
    InstructionDeclaration,
    InstructionMissing,
)

__all__ = ("InstructionsVolatileStorage",)


@final
class InstructionsVolatileStorage:
    def __init__(
        self,
        instructions: Sequence[Instruction],
    ) -> None:
        self._storage: Mapping[str, Instruction] = {
            instruction.name: instruction for instruction in instructions
        }
        self._listing: Sequence[InstructionDeclaration] = [
            instruction.declaration for instruction in instructions
        ]

    async def listing(
        self,
        **extra: Any,
    ) -> Sequence[InstructionDeclaration]:
        return self._listing

    async def instruction(
        self,
        name: str,
        *,
        arguments: Mapping[str, str | float | int] | None = None,
        **extra: Any,
    ) -> Instruction:
        if instruction := self._storage.get(name):
            if arguments:
                return instruction.updated(arguments={**instruction.arguments, **arguments})

            else:
                return instruction

        else:
            raise InstructionMissing(f"Instruction '{name}' is not defined")



from draive.vllm.client import VLLM
from draive.vllm.config import (
    VLLMChatConfig,
    VLLMEmbeddingConfig,
)
from draive.vllm.types import VLLMException

__all__ = (
    "VLLM",
    "VLLMChatConfig",
    "VLLMEmbeddingConfig",
    "VLLMException",
)



from asyncio import Task, gather
from collections.abc import AsyncIterable, AsyncIterator, Iterable
from typing import Any

from haiway import AsyncQueue, ctx

from draive.instructions import Instruction
from draive.lmm import (
    LMM,
    LMMCompletion,
    LMMContextElement,
    LMMInput,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolResponse,
    LMMToolResponses,
)
from draive.lmm.types import LMMStreamChunk
from draive.multimodal import Multimodal, MultimodalContent
from draive.prompts import Prompt
from draive.tools import Tool, Toolbox

__all__ = ("generate_text",)


async def generate_text(
    *,
    instruction: Instruction | str | None,
    input: Prompt | Multimodal,  # noqa: A002
    tools: Toolbox | Iterable[Tool] | None,
    examples: Iterable[tuple[Multimodal, str]] | None,
    stream: bool,
    **extra: Any,
) -> AsyncIterable[str] | str:
    with ctx.scope("generate_text"):
        toolbox: Toolbox = Toolbox.of(tools)

        context: list[LMMContextElement] = [
            *[
                message
                for example in examples or []
                for message in [
                    LMMInput.of(example[0]),
                    LMMCompletion.of(example[1]),
                ]
            ],
        ]

        match input:
            case Prompt() as prompt:
                context.extend(prompt.content)

            case value:
                context.append(LMMInput.of(value))

        if stream:
            return await _text_generation_stream(
                instruction=instruction,
                context=context,
                toolbox=toolbox,
            )

        else:
            return await _text_generation(
                instruction=instruction,
                context=context,
                toolbox=toolbox,
            )


async def _text_generation(
    instruction: Instruction | str | None,
    context: list[LMMContextElement],
    toolbox: Toolbox,
    **extra: Any,
) -> str:
    recursion_level: int = 0
    while recursion_level <= toolbox.repeated_calls_limit:
        match await LMM.completion(
            instruction=instruction,
            context=context,
            tools=toolbox.available_tools(),
            tool_selection=toolbox.tool_selection(repetition_level=recursion_level),
            output="text",
            **extra,
        ):
            case LMMCompletion() as completion:
                ctx.log_debug("Received text generation result")
                return completion.content.to_str()

            case LMMToolRequests() as tool_requests:
                ctx.log_debug("Received text generation tool calls")
                tool_responses: LMMToolResponses = await toolbox.respond_all(tool_requests)

                if direct_responses := [
                    response
                    for response in tool_responses.responses
                    if response.handling == "direct_result"
                ]:
                    return MultimodalContent.of(
                        *[response.content for response in direct_responses]
                    ).to_str()

                else:
                    context.extend(
                        [
                            tool_requests,
                            tool_responses,
                        ]
                    )

        recursion_level += 1  # continue with next recursion level

    raise RuntimeError("LMM exceeded limit of recursive calls")


async def _text_generation_stream(
    instruction: Instruction | str | None,
    context: list[LMMContextElement],
    toolbox: Toolbox,
    **extra: Any,
) -> AsyncIterator[str]:
    output_queue = AsyncQueue[str]()

    async def consume_lmm_output() -> None:
        recursion_level: int = 0
        try:
            while recursion_level <= toolbox.repeated_calls_limit:
                pending_tool_requests: list[LMMToolRequest] = []
                pending_tool_responses: list[Task[LMMToolResponse]] = []
                accumulated_text: str = ""

                async for element in await LMM.completion(
                    instruction=instruction,
                    context=context,
                    tool_selection=toolbox.tool_selection(repetition_level=recursion_level),
                    tools=toolbox.available_tools(),
                    output="text",
                    stream=True,
                    **extra,
                ):
                    match element:
                        case LMMStreamChunk() as chunk:
                            chunk_text: str = chunk.content.to_str()
                            accumulated_text += chunk_text
                            output_queue.enqueue(chunk_text)

                            if chunk.eod:
                                # end of streaming for text generation
                                return output_queue.finish()

                        case LMMToolRequest() as tool_request:
                            # we could start processing immediately
                            pending_tool_requests.append(tool_request)
                            pending_tool_responses.append(ctx.spawn(toolbox.respond, tool_request))

                assert pending_tool_requests  # nosec: B101
                tool_requests: LMMToolRequests = LMMToolRequests.of(
                    pending_tool_requests,
                    content=MultimodalContent.of(accumulated_text) if accumulated_text else None,
                )
                tool_responses: LMMToolResponses = LMMToolResponses.of(
                    await gather(*pending_tool_responses),
                )
                if direct_content := [
                    response.content
                    for response in tool_responses.responses
                    if response.handling == "direct_result"
                ]:
                    response_text: str = MultimodalContent.of(*direct_content).to_str()
                    output_queue.enqueue(response_text)
                    return output_queue.finish()

                else:
                    context.extend(
                        [
                            tool_requests,
                            tool_responses,
                        ]
                    )

                recursion_level += 1  # continue with next recursion level

            raise RuntimeError("LMM exceeded limit of tool calls")

        except BaseException as exc:
            output_queue.finish(exception=exc)
            raise exc

    ctx.spawn(consume_lmm_output)

    return output_queue



from draive.choice.call import choice_completion
from draive.choice.default import default_choice_completion
from draive.choice.state import Choice
from draive.choice.types import ChoiceCompletion, ChoiceOption, SelectionException

__all__ = (
    "Choice",
    "ChoiceCompletion",
    "ChoiceOption",
    "SelectionException",
    "choice_completion",
    "default_choice_completion",
)



from asyncio import Lock, gather
from collections.abc import Callable, Iterable, Sequence
from pathlib import Path
from typing import Literal, Protocol, Self, overload, runtime_checkable
from uuid import UUID, uuid4

from haiway import ScopeContext, asynchronous, ctx

from draive.commons import META_EMPTY, Meta, MetaValues
from draive.evaluation.evaluator import EvaluatorResult, PreparedEvaluator
from draive.evaluation.generator import generate_case_parameters
from draive.evaluation.scenario import PreparedScenarioEvaluator, ScenarioEvaluatorResult
from draive.parameters import DataModel, Field

__all__ = (
    "EvaluationCaseResult",
    "EvaluationSuite",
    "EvaluationSuiteDefinition",
    "EvaluationSuiteStorage",
    "SuiteEvaluatorCaseResult",
    "SuiteEvaluatorResult",
    "evaluation_suite",
)


class EvaluationSuiteCase[CaseParameters: DataModel](DataModel):
    identifier: UUID = Field(default_factory=uuid4)
    parameters: CaseParameters
    comment: str | None = None


class SuiteEvaluatorCaseResult[CaseParameters: DataModel, Value: DataModel | str](DataModel):
    case: EvaluationSuiteCase[CaseParameters] = Field(
        description="Evaluated case",
    )
    value: Value = Field(
        description="Evaluated value",
    )
    results: Sequence[ScenarioEvaluatorResult] = Field(
        description="Evaluation results",
    )
    meta: Meta = Field(
        description="Additional evaluation metadata",
        default=META_EMPTY,
    )

    @property
    def passed(self) -> bool:
        # empty results is equivalent of failure
        return len(self.results) > 0 and all(result.passed for result in self.results)

    def report(
        self,
        *,
        include_passed: bool = True,
        include_details: bool = True,
    ) -> str:
        report: str = "\n".join(
            [
                result.report(
                    include_passed=include_passed,
                    include_details=include_details,
                )
                for result in self.results
                if include_passed or not result.passed
            ]
        )

        if report:  # nonempty report
            if include_details:
                return (
                    f"<evaluation_case identifier='{self.case.identifier}'>"
                    f"\n<relative_score>{self.relative_score * 100:.2f}%</relative_score>"
                    f"\n<evaluated_value>{self.value}</evaluated_value>"
                    # TODO: convert DataModel to xml representation when avaialble
                    f"\n<parameters>{self.case.parameters}</parameters>"
                    f"\n{report}"
                    "\n</evaluation_case>"
                )

            else:
                return f"Evaluation case {self.case.identifier}:\n{report}\n---\n"

        elif not self.results:
            return f"Evaluation case {self.case.identifier} empty!"

        else:
            return f"Evaluation case {self.case.identifier} passed!"

    @property
    def relative_score(self) -> float:
        if not self.results:
            return 0.0

        score: float = 0.0
        for evaluation in self.results:
            score += evaluation.relative_score

        return score / len(self.results)


class SuiteEvaluatorResult[CaseParameters: DataModel, Value: DataModel | str](DataModel):
    cases: Sequence[SuiteEvaluatorCaseResult[CaseParameters, Value]]

    @property
    def passed(self) -> bool:
        return all(case.passed for case in self.cases)

    def report(
        self,
        *,
        include_passed: bool = True,
        include_details: bool = True,
    ) -> str:
        report: str = "\n".join(
            [
                result.report(
                    include_passed=include_passed,
                    include_details=include_details,
                )
                for result in self.cases
                if include_passed or not result.passed
            ]
        )

        if report:  # nonempty report contains failing reports
            return f"Evaluation suite failed:\n\n{report}"

        elif not self.cases:
            return "Evaluation suite empty!"

        else:
            return "Evaluation suite passed!"

    @property
    def relative_score(self) -> float:
        if not self.cases:
            return 0.0

        score: float = 0.0
        for evaluation in self.cases:
            score += evaluation.relative_score

        return score / len(self.cases)


class EvaluationCaseResult[Value: DataModel | str](DataModel):
    @classmethod
    def of(
        cls,
        results: ScenarioEvaluatorResult | EvaluatorResult,
        *_results: ScenarioEvaluatorResult | EvaluatorResult,
        value: Value,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        free_results: list[EvaluatorResult] = []
        scenario_results: list[ScenarioEvaluatorResult] = []
        for result in (results, *_results):
            match result:
                case ScenarioEvaluatorResult() as scenario_result:
                    scenario_results.append(scenario_result)

                case EvaluatorResult() as evaluator_result:
                    free_results.append(evaluator_result)

        if free_results:
            scenario_results.append(
                ScenarioEvaluatorResult(
                    scenario="EvaluationSuite",
                    evaluations=tuple(free_results),
                )
            )

        return cls(
            value=value,
            results=tuple(scenario_results),
            meta=Meta.of(meta),
        )

    @classmethod
    async def evaluating(
        cls,
        value: Value,
        /,
        evaluators: PreparedScenarioEvaluator[Value] | PreparedEvaluator[Value],
        *_evaluators: PreparedScenarioEvaluator[Value] | PreparedEvaluator[Value],
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        return cls.of(
            *await gather(
                *[evaluator(value) for evaluator in [evaluators, *_evaluators]],
                return_exceptions=False,
            ),
            value=value,
            meta=Meta.of(meta),
        )

    value: Value = Field(
        description="Evaluated value",
    )
    results: Sequence[ScenarioEvaluatorResult] = Field(
        description="Evaluation results",
    )
    meta: Meta = Field(
        description="Additional evaluation metadata",
        default=META_EMPTY,
    )


@runtime_checkable
class EvaluationSuiteDefinition[CaseParameters: DataModel, Value: DataModel | str](Protocol):
    async def __call__(
        self,
        parameters: CaseParameters,
    ) -> EvaluationCaseResult[Value]: ...


class EvaluationSuiteData[CaseParameters: DataModel](DataModel):
    cases: Sequence[EvaluationSuiteCase[CaseParameters]] = Field(default_factory=tuple)


@runtime_checkable
class EvaluationSuiteStorage[CaseParameters: DataModel](Protocol):
    async def load(
        self,
    ) -> EvaluationSuiteData[CaseParameters]: ...

    async def save(
        self,
        data: EvaluationSuiteData[CaseParameters],
    ) -> None: ...


class EvaluationSuite[CaseParameters: DataModel, Value: DataModel | str]:
    def __init__(
        self,
        parameters: type[CaseParameters],
        definition: EvaluationSuiteDefinition[CaseParameters, Value],
        storage: EvaluationSuiteStorage[CaseParameters],
        execution_context: ScopeContext | None,
    ) -> None:
        self._parameters: type[CaseParameters] = parameters
        self._definition: EvaluationSuiteDefinition[CaseParameters, Value] = definition
        self._storage: EvaluationSuiteStorage[CaseParameters] = storage
        self._execution_context: ScopeContext | None = execution_context
        self._data_cache: EvaluationSuiteData[CaseParameters] | None = None
        self._lock: Lock = Lock()

    @overload
    async def __call__(
        self,
        parameters: CaseParameters,
        /,
        *,
        reload: Literal[False] = False,
        store: Literal[True],
    ) -> SuiteEvaluatorCaseResult[CaseParameters, Value]: ...

    @overload
    async def __call__(
        self,
        parameters: CaseParameters | UUID,
        /,
        *,
        reload: bool = False,
        store: Literal[False] = False,
    ) -> SuiteEvaluatorCaseResult[CaseParameters, Value]: ...

    @overload
    async def __call__(
        self,
        /,
        *,
        reload: bool = False,
        store: Literal[False] = False,
    ) -> SuiteEvaluatorResult[CaseParameters, Value]: ...

    async def __call__(
        self,
        parameters: CaseParameters | UUID | None = None,
        /,
        *,
        reload: bool = False,
        store: bool = False,
    ) -> (
        SuiteEvaluatorResult[CaseParameters, Value]
        | SuiteEvaluatorCaseResult[CaseParameters, Value]
    ):
        if context := self._execution_context:
            async with context:
                return await self._evaluate(
                    parameters,
                    reload=reload,
                )

        else:
            async with ctx.scope("evaluation.suite"):
                return await self._evaluate(
                    parameters,
                    reload=reload,
                )

    async def _evaluate(
        self,
        parameters: CaseParameters | UUID | None = None,
        /,
        *,
        reload: bool = False,
    ) -> (
        SuiteEvaluatorResult[CaseParameters, Value]
        | SuiteEvaluatorCaseResult[CaseParameters, Value]
    ):
        async with self._lock:
            match parameters:
                case None:
                    return SuiteEvaluatorResult(
                        cases=await gather(
                            *[
                                self._evaluate_case(case=case)
                                for case in (await self._data(reload=reload)).cases
                            ],
                            return_exceptions=False,
                        )
                    )

                case UUID() as identifier:
                    available_cases: Sequence[EvaluationSuiteCase[CaseParameters]] = (
                        await self._data(reload=reload)
                    ).cases

                    if evaluation_case := next(
                        iter([case for case in available_cases if case.identifier == identifier]),
                        None,
                    ):
                        return await self._evaluate_case(case=evaluation_case)

                    else:
                        raise ValueError(f"Evaluation case with ID {identifier} does not exists.")

                case case_parameters:
                    return await self._evaluate_case(
                        case=EvaluationSuiteCase[CaseParameters](
                            parameters=case_parameters,
                        )
                    )

    async def _evaluate_case(
        self,
        *,
        case: EvaluationSuiteCase[CaseParameters],
    ) -> SuiteEvaluatorCaseResult[CaseParameters, Value]:
        result: EvaluationCaseResult[Value] = await self._definition(parameters=case.parameters)

        return SuiteEvaluatorCaseResult[CaseParameters, Value](
            case=case,
            value=result.value,
            results=result.results,
        )

    async def _data(
        self,
        reload: bool = False,
    ) -> EvaluationSuiteData[CaseParameters]:
        if reload or self._data_cache is None:
            self._data_cache = await self._storage.load()
            return self._data_cache

        else:
            return self._data_cache

    def with_execution_context(
        self,
        context: ScopeContext,
        /,
    ) -> Self:
        return self.__class__(
            parameters=self._parameters,
            definition=self._definition,
            storage=self._storage,
            execution_context=context,
        )

    def with_storage(
        self,
        storage: EvaluationSuiteStorage[CaseParameters] | Sequence[CaseParameters] | Path | str,
        /,
    ) -> Self:
        suite_storage: EvaluationSuiteStorage[CaseParameters]
        match storage:
            case str() as path_str:
                suite_storage = _EvaluationSuiteFileStorage[CaseParameters](
                    path=path_str,
                    data_type=EvaluationSuiteData[self._parameters],
                )

            case Path() as path:
                suite_storage = _EvaluationSuiteFileStorage[CaseParameters](
                    path=path,
                    data_type=EvaluationSuiteData[self._parameters],
                )

            case [*cases]:
                suite_storage = _EvaluationSuiteMemoryStorage[CaseParameters](
                    data_type=EvaluationSuiteData[CaseParameters],
                    cases=[EvaluationSuiteCase(parameters=parameters) for parameters in cases],
                )

            case storage:
                suite_storage = storage

        return self.__class__(
            parameters=self._parameters,
            definition=self._definition,
            storage=suite_storage,
            execution_context=self._execution_context,
        )

    async def cases(
        self,
        reload: bool = False,
    ) -> Sequence[EvaluationSuiteCase[CaseParameters]]:
        async with self._lock:
            return (await self._data(reload=reload)).cases

    async def add_case(
        self,
        parameters: CaseParameters,
        /,
        comment: str | None = None,
    ) -> None:
        async with self._lock:
            data: EvaluationSuiteData[CaseParameters] = await self._data(reload=True)
            self._data_cache = data.updated(
                cases=(
                    *data.cases,
                    EvaluationSuiteCase[CaseParameters](
                        parameters=parameters,
                        comment=comment,
                    ),
                )
            )
            await self._storage.save(self._data_cache)

    async def generate_cases(
        self,
        *,
        count: int,
        persist: bool = False,
        guidelines: str | None = None,
        examples: Iterable[CaseParameters] | None = None,
    ) -> Sequence[EvaluationSuiteCase[CaseParameters]]:
        async with self._lock:
            current: EvaluationSuiteData[CaseParameters] = await self._data(reload=True)

            generated: Sequence[EvaluationSuiteCase[CaseParameters]] = [
                EvaluationSuiteCase(parameters=parameters)
                for parameters in await generate_case_parameters(
                    self._parameters,
                    count=count,
                    examples=examples or [case.parameters for case in current.cases],
                    guidelines=guidelines,
                )
            ]

            self._data_cache = current.updated(cases=(*current.cases, *generated))

            if persist:
                await self._storage.save(self._data_cache)

            return generated

    async def remove_case(
        self,
        identifier: UUID,
        /,
    ) -> None:
        async with self._lock:
            data: EvaluationSuiteData[CaseParameters] = await self._data(reload=True)
            self._data_cache = data.updated(
                cases=tuple(case for case in data.cases if case.identifier != identifier)
            )
            await self._storage.save(self._data_cache)


def evaluation_suite[CaseParameters: DataModel, Value: DataModel | str](
    case: type[CaseParameters],
    /,
    storage: EvaluationSuiteStorage[CaseParameters]
    | Sequence[CaseParameters]
    | Path
    | str
    | None = None,
    execution_context: ScopeContext | None = None,
) -> Callable[
    [EvaluationSuiteDefinition[CaseParameters, Value]],
    EvaluationSuite[CaseParameters, Value],
]:
    suite_storage: EvaluationSuiteStorage[CaseParameters]
    match storage:
        case None:
            suite_storage = _EvaluationSuiteMemoryStorage[CaseParameters](
                data_type=EvaluationSuiteData[case],
                cases=(),
            )

        case str() as path_str:
            suite_storage = _EvaluationSuiteFileStorage[CaseParameters](
                path=path_str,
                data_type=EvaluationSuiteData[case],
            )

        case Path() as path:
            suite_storage = _EvaluationSuiteFileStorage[CaseParameters](
                path=path,
                data_type=EvaluationSuiteData[case],
            )

        case [*cases]:
            suite_storage = _EvaluationSuiteMemoryStorage[CaseParameters](
                data_type=EvaluationSuiteData[case],
                cases=[EvaluationSuiteCase(parameters=parameters) for parameters in cases],
            )

        case storage:
            suite_storage = storage

    def wrap(
        definition: EvaluationSuiteDefinition[CaseParameters, Value],
    ) -> EvaluationSuite[CaseParameters, Value]:
        return EvaluationSuite[CaseParameters, Value](
            parameters=case,
            definition=definition,
            storage=suite_storage,
            execution_context=execution_context,
        )

    return wrap


class _EvaluationSuiteMemoryStorage[CaseParameters: DataModel]:
    def __init__(
        self,
        data_type: type[EvaluationSuiteData[CaseParameters]],
        cases: Sequence[EvaluationSuiteCase[CaseParameters]] | None,
    ) -> None:
        self.store: Sequence[EvaluationSuiteCase[CaseParameters]] = cases or ()
        self._data_type: type[EvaluationSuiteData[CaseParameters]] = data_type

    async def load(
        self,
    ) -> EvaluationSuiteData[CaseParameters]:
        return self._data_type(cases=self.store)

    async def save(
        self,
        data: EvaluationSuiteData[CaseParameters],
    ) -> None:
        self.store = data.cases


class _EvaluationSuiteFileStorage[CaseParameters: DataModel]:
    def __init__(
        self,
        path: Path | str,
        data_type: type[EvaluationSuiteData[CaseParameters]],
    ) -> None:
        self._path: Path
        match path:
            case str() as path_str:
                self._path = Path(path_str)

            case path:
                self._path = path

        self._data_type: type[EvaluationSuiteData[CaseParameters]] = data_type

    async def load(
        self,
    ) -> EvaluationSuiteData[CaseParameters]:
        try:
            return await self._file_load(data_type=self._data_type)

        except ValueError as exc:
            ctx.log_error(
                f"Invalid EvaluationSuite at {self._path}",
                exception=exc,
            )
            return self._data_type()

    @asynchronous
    def _file_load(
        self,
        data_type: type[EvaluationSuiteData[CaseParameters]],
    ) -> EvaluationSuiteData[CaseParameters]:
        if self._path.exists():
            with open(self._path, mode="rb") as file:
                return data_type.from_json(file.read())

        else:
            return data_type()

    async def save(
        self,
        data: EvaluationSuiteData[CaseParameters],
    ) -> None:
        await self._file_save(data=data)

    @asynchronous
    def _file_save(
        self,
        data: EvaluationSuiteData[CaseParameters],
    ) -> None:
        with open(self._path, mode="wb+") as file:
            file.write(data.to_json(indent=2).encode("utf-8"))



from typing import Any, cast

from haiway import asynchronous, cache
from sentencepiece import SentencePieceProcessor  # pyright: ignore[reportMissingTypeStubs]

from draive.tokenization import TextTokenizing

__all__ = (
    "sentencepiece_processor",
    "sentencepiece_tokenizer",
)


def sentencepiece_tokenizer(
    processor: SentencePieceProcessor,
    /,
) -> TextTokenizing:
    def sentencepiece_tokenize_text(
        text: str,
        **extra: Any,
    ) -> list[int]:
        return cast(
            list[int],
            processor.encode(text),  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue]
        )

    return sentencepiece_tokenize_text


@asynchronous
@cache(limit=1)
def sentencepiece_processor(
    *,
    model_path: str,
) -> SentencePieceProcessor:
    return SentencePieceProcessor(model_file=model_path)  # pyright: ignore[reportCallIssue]



from typing import Any, Protocol, runtime_checkable

from draive.instructions import Instruction
from draive.multimodal import MediaContent, Multimodal

__all__ = ("ImageGenerating",)


@runtime_checkable
class ImageGenerating(Protocol):
    async def __call__(
        self,
        *,
        instruction: Instruction | str,
        input: Multimodal | None,  # noqa: A002
        **extra: Any,
    ) -> MediaContent: ...



from asyncio import gather
from base64 import urlsafe_b64encode
from collections.abc import Callable, Sequence
from itertools import chain
from typing import Any, cast

from cohere import EmbedByTypeResponse
from haiway import ObservabilityLevel, State, as_list, ctx

from draive.cohere.api import CohereAPI
from draive.cohere.config import CohereImageEmbeddingConfig, CohereTextEmbeddingConfig
from draive.embedding import Embedded, ImageEmbedding, TextEmbedding
from draive.parameters import DataModel

__all__ = ("CohereEmbedding",)


class CohereEmbedding(CohereAPI):
    def text_embedding(self) -> TextEmbedding:
        """
        Prepare TextEmbedding implementation using Cohere service.
        """
        return TextEmbedding(embedding=self.create_texts_embedding)

    async def create_texts_embedding[Value: DataModel | State](
        self,
        values: Sequence[Value] | Sequence[str],
        /,
        attribute: Callable[[Value], str] | None = None,
        *,
        config: CohereTextEmbeddingConfig | None = None,
        **extra: Any,
    ) -> Sequence[Embedded[Value]] | Sequence[Embedded[str]]:
        """
        Create texts embedding with Cohere embedding service.
        """
        embedding_config: CohereTextEmbeddingConfig = config or ctx.state(CohereTextEmbeddingConfig)
        with ctx.scope("cohere_text_embedding", embedding_config):
            ctx.record(
                ObservabilityLevel.INFO,
                attributes={
                    "embedding.model": embedding_config.model,
                    "embedding.purpose": embedding_config.purpose,
                },
            )

            attributes: list[str]
            if attribute is None:
                attributes = cast(list[str], as_list(values))

            else:
                attributes = [attribute(cast(Value, value)) for value in values]

            assert all(isinstance(element, str) for element in attributes)  # nosec: B101

            responses: list[EmbedByTypeResponse] = await gather(
                *[
                    self._client.embed(
                        model=embedding_config.model,
                        texts=attributes[index : index + embedding_config.batch_size],
                        embedding_types=["float"],
                        input_type=embedding_config.purpose,
                    )
                    for index in range(0, len(attributes), embedding_config.batch_size)
                ]
            )

            return cast(
                Sequence[Embedded[Value]] | Sequence[Embedded[str]],
                [
                    Embedded(
                        value=value,
                        vector=embedding,
                    )
                    for value, embedding in zip(
                        values,
                        chain.from_iterable(
                            [
                                cast(list[list[float]], response.embeddings.float_)
                                for response in responses
                            ]
                        ),
                        strict=True,
                    )
                ],
            )

    def image_embedding(self) -> ImageEmbedding:
        """
        Prepare ImageEmbedding implementation using Cohere service.
        """
        return ImageEmbedding(embedding=self.create_images_embedding)

    async def create_images_embedding[Value: DataModel | State](
        self,
        values: Sequence[Value] | Sequence[bytes],
        /,
        attribute: Callable[[Value], bytes] | None = None,
        *,
        config: CohereImageEmbeddingConfig | None = None,
        **extra: Any,
    ) -> Sequence[Embedded[Value]] | Sequence[Embedded[bytes]]:
        """
        Create image embedding with Cohere embedding service.
        """
        embedding_config: CohereImageEmbeddingConfig = config or ctx.state(
            CohereImageEmbeddingConfig
        )
        with ctx.scope("cohere_image_embedding", embedding_config):
            ctx.record(
                ObservabilityLevel.INFO,
                attributes={
                    "embedding.model": embedding_config.model,
                },
            )
            attributes: list[bytes]
            if attribute is None:
                attributes = cast(list[bytes], as_list(values))

            else:
                attributes = [attribute(cast(Value, value)) for value in values]

            assert all(isinstance(element, bytes) for element in attributes)  # nosec: B101

            responses: list[EmbedByTypeResponse] = await gather(
                *[
                    self._client.embed(
                        model=embedding_config.model,
                        images=[
                            f"data:image/jpeg;base64,{urlsafe_b64encode(image).decode('utf-8')}"
                            for image in attributes[index : index + embedding_config.batch_size]
                        ],
                        embedding_types=["float"],
                        input_type="image",
                    )
                    for index in range(0, len(attributes), embedding_config.batch_size)
                ]
            )

            return cast(
                Sequence[Embedded[Value]] | Sequence[Embedded[bytes]],
                [
                    Embedded(
                        value=value,
                        vector=embedding,
                    )
                    for value, embedding in zip(
                        values,
                        chain.from_iterable(
                            [
                                cast(list[list[float]], response.embeddings.float_)
                                for response in responses
                            ]
                        ),
                        strict=True,
                    )
                ],
            )



import base64
import json
import re
import typing

import httpx
from cohere import (
    ApiMeta,
    ApiMetaBilledUnits,
    ApiMetaTokens,
    EmbedResponse,
    GenerateStreamedResponse,
    Generation,
    NonStreamedChatResponse,
    RerankResponse,
    StreamedChatResponse,
)
from cohere.client import ClientEnvironment  # pyright: ignore
from cohere.client_v2 import AsyncClientV2
from cohere.core import construct_type
from cohere.manually_maintained.lazy_aws_deps import lazy_boto3, lazy_botocore
from haiway import asynchronous
from httpx import URL, AsyncByteStream, ByteStream

__all__ = ("AsyncBedrockClientV2",)


# it is a naive adjustment of cohere.BedrockClientV2 to work with async interfaces
class AsyncAwsClientV2(AsyncClientV2):
    def __init__(
        self,
        *,
        aws_region: str | None = None,
        timeout: float | None = None,
        service: typing.Literal["bedrock"] | typing.Literal["sagemaker"],
    ):
        AsyncClientV2.__init__(
            self,
            base_url="https://api.cohere.com",  # this url is unused for BedrockClient
            environment=ClientEnvironment.PRODUCTION,
            client_name="n/a",
            timeout=timeout,
            api_key="n/a",
            httpx_client=httpx.AsyncClient(
                event_hooks=get_event_hooks(
                    service=service,
                    aws_region=aws_region,
                ),
                timeout=timeout,
            ),
        )


class AsyncBedrockClientV2(AsyncAwsClientV2):
    def __init__(
        self,
        *,
        aws_region: str | None = None,
        timeout: float | None = None,
    ):
        AsyncAwsClientV2.__init__(
            self,
            service="bedrock",
            aws_region=aws_region,
            timeout=timeout,
        )


EventHook = typing.Callable[..., typing.Any]


def get_event_hooks(
    service: str,
    aws_region: str | None,
) -> dict[str, list[EventHook]]:
    return {
        "request": [
            map_request_to_bedrock(
                service=service,
                aws_region=aws_region,
            ),
        ],
        "response": [map_response_from_bedrock()],
    }


class TextGeneration(typing.TypedDict):
    text: str
    is_finished: str
    event_type: typing.Literal["text-generation"]


class StreamEnd(typing.TypedDict):
    is_finished: str
    event_type: typing.Literal["stream-end"]
    finish_reason: str


class Streamer(AsyncByteStream):
    lines: typing.AsyncIterator[bytes]

    def __init__(self, lines: typing.AsyncIterator[bytes]):
        self.lines = lines

    def __aiter__(self) -> typing.AsyncIterator[bytes]:
        return self.lines


response_mapping: dict[str, typing.Any] = {
    "chat": NonStreamedChatResponse,
    "embed": EmbedResponse,
    "generate": Generation,
    "rerank": RerankResponse,
}

stream_response_mapping: dict[str, typing.Any] = {
    "chat": StreamedChatResponse,
    "generate": GenerateStreamedResponse,
}


async def stream_generator(response: httpx.Response, endpoint: str) -> typing.AsyncIterator[bytes]:
    regex = r"{[^\}]*}"

    async for _text in response.aiter_lines():
        match = re.search(regex, _text)
        if match:
            obj = json.loads(match.group())
            if "bytes" in obj:
                base64_payload = base64.b64decode(obj["bytes"]).decode("utf-8")
                streamed_obj = json.loads(base64_payload)
                if "event_type" in streamed_obj:
                    response_type = stream_response_mapping[endpoint]
                    parsed = typing.cast(
                        response_type,  # type: ignore
                        construct_type(type_=response_type, object_=streamed_obj),
                    )
                    yield (json.dumps(parsed.dict()) + "\n").encode("utf-8")  # type: ignore


def map_token_counts(response: httpx.Response) -> ApiMeta:
    input_tokens = int(response.headers.get("X-Amzn-Bedrock-Input-Token-Count", -1))
    output_tokens = int(response.headers.get("X-Amzn-Bedrock-Output-Token-Count", -1))
    return ApiMeta(
        tokens=ApiMetaTokens(input_tokens=input_tokens, output_tokens=output_tokens),
        billed_units=ApiMetaBilledUnits(input_tokens=input_tokens, output_tokens=output_tokens),
    )


def map_response_from_bedrock():
    async def _hook(
        response: httpx.Response,
    ) -> None:
        stream = response.headers["content-type"] == "application/vnd.amazon.eventstream"
        endpoint = response.request.extensions["endpoint"]
        output: typing.AsyncIterator[bytes]

        if stream:
            output = stream_generator(
                httpx.Response(
                    stream=response.stream,
                    status_code=response.status_code,
                ),
                endpoint,
            )

        else:
            response_type = response_mapping[endpoint]
            response_obj = json.loads(await response.aread())
            response_obj["meta"] = map_token_counts(response).dict()
            cast_obj: typing.Any = typing.cast(
                response_type,  # type: ignore
                construct_type(
                    type_=response_type,
                    # type: ignore
                    object_=response_obj,
                ),
            )

            async def output_generator() -> typing.AsyncIterator[bytes]:
                yield json.dumps(cast_obj.dict()).encode("utf-8")

            output = output_generator()

        response.stream = Streamer(output)

        # reset response object to allow for re-reading
        if hasattr(response, "_content"):
            del response._content
        response.is_stream_consumed = False
        response.is_closed = False

    return _hook


def map_request_to_bedrock(
    service: str,
    aws_region: str | None,
) -> EventHook:
    session = lazy_boto3().Session(region_name=aws_region)
    aws_region = session.region_name
    credentials = session.get_credentials()
    signer = lazy_botocore().auth.SigV4Auth(credentials, service, aws_region)  # pyright: ignore

    @asynchronous
    def _event_hook(request: httpx.Request) -> None:
        headers = request.headers.copy()
        del headers["connection"]

        api_version = request.url.path.split("/")[-2]
        endpoint = request.url.path.split("/")[-1]
        body = json.loads(request.read())
        model = body["model"]

        url = get_url(
            platform=service,
            aws_region=aws_region,
            model=model,  # type: ignore
            stream="stream" in body and body["stream"],
        )
        request.url = URL(url)
        request.headers["host"] = request.url.host

        if endpoint == "rerank":
            body["api_version"] = get_api_version(version=api_version)

        if "stream" in body:
            del body["stream"]

        if "model" in body:
            del body["model"]

        new_body = json.dumps(body).encode("utf-8")
        request.stream = ByteStream(new_body)
        request._content = new_body
        headers["content-length"] = str(len(new_body))

        aws_request = lazy_botocore().awsrequest.AWSRequest(  # pyright: ignore
            method=request.method,
            url=url,
            headers=headers,
            data=request.read(),
        )
        signer.add_auth(aws_request)

        request.headers = httpx.Headers(aws_request.prepare().headers)
        request.extensions["endpoint"] = endpoint

    return _event_hook


def get_url(
    *,
    platform: str,
    aws_region: str | None,
    model: str,
    stream: bool,
) -> str:
    if platform == "bedrock":
        endpoint = "invoke" if not stream else "invoke-with-response-stream"
        return f"https://{platform}-runtime.{aws_region}.amazonaws.com/model/{model}/{endpoint}"
    elif platform == "sagemaker":
        endpoint = "invocations" if not stream else "invocations-response-stream"
        return f"https://runtime.sagemaker.{aws_region}.amazonaws.com/endpoints/{model}/{endpoint}"
    return ""


def get_api_version(*, version: str):
    int_version = {
        "v1": 1,
        "v2": 2,
    }

    return int_version.get(version, 1)



from typing import Any

from haiway import ObservabilityLevel, ctx, not_missing
from openai.types import ModerationCreateResponse, ModerationMultiModalInputParam

from draive.guardrails import GuardrailsModeration, GuardrailsModerationException
from draive.multimodal import MediaData, MediaReference, Multimodal, MultimodalContent, TextContent
from draive.openai.api import OpenAIAPI
from draive.openai.config import OpenAIModerationConfig
from draive.openai.utils import unwrap_missing

__all__ = ("OpenAIContentModereation",)


class OpenAIContentModereation(OpenAIAPI):
    def content_guardrails(self) -> GuardrailsModeration:
        return GuardrailsModeration(input_checking=self.content_verification)

    async def content_verification(  # noqa: C901, PLR0912, PLR0915
        self,
        content: Multimodal,
        /,
        *,
        config: OpenAIModerationConfig | None = None,
        **extra: Any,
    ) -> None:
        moderation_config: OpenAIModerationConfig = config or ctx.state(OpenAIModerationConfig)
        with ctx.scope("openai_moderation", moderation_config):
            ctx.record(
                ObservabilityLevel.INFO,
                attributes={
                    "guardrails.provider": "openai",
                    "guardrails.model": moderation_config.model,
                },
            )
            content = MultimodalContent.of(content)
            moderated_content: list[ModerationMultiModalInputParam] = []
            for part in content.parts:
                match part:
                    case TextContent() as text:
                        moderated_content.append(
                            {
                                "type": "text",
                                "text": text.text,
                            }
                        )

                    case MediaData() as media_data:
                        moderated_content.append(
                            {
                                "type": "image_url",
                                "image_url": {
                                    "url": media_data.to_data_uri(safe_encoding=False),
                                },
                            }
                        )

                    case MediaReference() as media_referencve:
                        moderated_content.append(
                            {
                                "type": "image_url",
                                "image_url": {"url": media_referencve.uri},
                            }
                        )

                    case other:
                        ctx.log_warning(
                            f"Attempting to use moderation on unsupported content ({type(other)}),"
                            " verifying as json text..."
                        )
                        moderated_content.append(
                            {
                                "type": "text",
                                "text": other.to_json(),
                            }
                        )

            response: ModerationCreateResponse = await self._client.moderations.create(
                model=moderation_config.model,
                input=moderated_content,
                timeout=unwrap_missing(moderation_config.timeout),
            )

            violations: set[str] = set()
            for result in response.results:
                if (
                    not_missing(moderation_config.harassment_threshold)
                    and result.category_scores.harassment >= moderation_config.harassment_threshold
                ):
                    violations.add("harassment")

                elif result.categories.harassment:
                    violations.add("harassment")

                if (
                    not_missing(moderation_config.harassment_threatening_threshold)
                    and result.category_scores.harassment_threatening
                    >= moderation_config.harassment_threatening_threshold
                ):
                    violations.add("harassment_threatening")

                elif result.categories.harassment_threatening:
                    violations.add("harassment_threatening")

                if (
                    not_missing(moderation_config.hate_threshold)
                    and result.category_scores.hate >= moderation_config.hate_threshold
                ):
                    violations.add("hate")

                elif result.categories.hate:
                    violations.add("hate")

                if (
                    not_missing(moderation_config.hate_threatening_threshold)
                    and result.category_scores.hate_threatening
                    >= moderation_config.hate_threatening_threshold
                ):
                    violations.add("hate_threatening")

                elif result.categories.hate_threatening:
                    violations.add("hate_threatening")

                if (
                    not_missing(moderation_config.self_harm_threshold)
                    and result.category_scores.self_harm >= moderation_config.self_harm_threshold
                ):
                    violations.add("self_harm")

                elif result.categories.self_harm:
                    violations.add("self_harm")

                if (
                    not_missing(moderation_config.self_harm_instructions_threshold)
                    and result.category_scores.self_harm_instructions
                    >= moderation_config.self_harm_instructions_threshold
                ):
                    violations.add("self_harm_instructions")

                elif result.categories.self_harm_instructions:
                    violations.add("self_harm_instructions")

                if (
                    not_missing(moderation_config.self_harm_intent_threshold)
                    and result.category_scores.self_harm_intent
                    >= moderation_config.self_harm_intent_threshold
                ):
                    violations.add("self_harm_intent")

                elif result.categories.self_harm_intent:
                    violations.add("self_harm_intent")

                if (
                    not_missing(moderation_config.sexual_threshold)
                    and result.category_scores.sexual >= moderation_config.sexual_threshold
                ):
                    violations.add("sexual")

                elif result.categories.sexual:
                    violations.add("sexual")

                if (
                    not_missing(moderation_config.sexual_minors_threshold)
                    and result.category_scores.sexual_minors
                    >= moderation_config.sexual_minors_threshold
                ):
                    violations.add("sexual_minors")

                elif result.categories.sexual_minors:
                    violations.add("sexual_minors")

                if (
                    not_missing(moderation_config.violence_threshold)
                    and result.category_scores.violence >= moderation_config.violence_threshold
                ):
                    violations.add("violence")

                elif result.categories.violence:
                    violations.add("violence")

                if (
                    not_missing(moderation_config.violence_graphic_threshold)
                    and result.category_scores.violence_graphic
                    >= moderation_config.violence_graphic_threshold
                ):
                    violations.add("violence_graphic")

                elif result.categories.violence_graphic:
                    violations.add("violence_graphic")

                if (
                    not_missing(moderation_config.illicit_threshold)
                    and result.category_scores.illicit >= moderation_config.illicit_threshold
                ):
                    violations.add("illicit")

                elif result.categories.illicit:
                    violations.add("illicit")

                if (
                    not_missing(moderation_config.illicit_violent_threshold)
                    and result.category_scores.illicit_violent
                    >= moderation_config.illicit_violent_threshold
                ):
                    violations.add("illicit_violent")

                elif result.categories.illicit_violent:
                    violations.add("illicit_violent")

            if violations:
                raise GuardrailsModerationException(
                    f"Content violated rule(s): {violations}",
                    violations=tuple(violations),
                    content=content,
                )



from asyncio import gather
from collections.abc import (
    Callable,
    Collection,
    Coroutine,
    Hashable,
    Iterable,
    Mapping,
    MutableMapping,
    Sequence,
)
from typing import Any, ClassVar, Literal, Protocol, Self, cast, final, overload
from uuid import uuid4

from haiway import Disposable, Disposables, State, cache, ctx, retry

from draive.commons import Meta, MetaValue, MetaValues
from draive.instructions import Instruction
from draive.lmm import (
    LMM,
    LMMCompletion,
    LMMContext,
    LMMContextElement,
    LMMInput,
    LMMOutputSelection,
    LMMToolRequests,
    LMMToolResponses,
)
from draive.multimodal import Multimodal, MultimodalContent
from draive.multimodal.tags import MultimodalTagElement
from draive.parameters import DataModel
from draive.prompts import Prompt
from draive.stages.types import (
    StageConditioning,
    StageContextTransforming,
    StageException,
    StageExecution,
    StageLoopConditioning,
    StageMemory,
    StageMerging,
    StageResultTransforming,
    StageRouting,
    StageState,
)
from draive.tools import Tool, Toolbox
from draive.utils.processing import Processing

__all__ = (
    "Stage",
    "stage",
)


class MakeCacheKey[Key](Protocol):
    """Protocol for generating cache keys from stage state."""

    def __call__(
        self,
        *,
        state: StageState,
    ) -> Key: ...


class ReadCache[Key](Protocol):
    """Protocol for reading cached stage states."""

    async def __call__(
        self,
        key: Key,
    ) -> StageState | None: ...


class WriteCache[Key](Protocol):
    """Protocol for writing stage states to cache."""

    async def __call__(
        self,
        key: Key,
        value: StageState,
    ) -> None: ...


@final
class Stage:
    """
    A Stage represents a unit of processing within an LMM pipeline.

    Stages are immutable execution units that process LMM context and produce results.
    They can be composed, cached, retried, and conditionally executed to build complex
    AI processing workflows.

    Key Features:
    - Immutable and composable design
    - Built-in caching, retry, and fallback mechanisms
    - Context and result transformation capabilities
    - Support for concurrent and sequential execution
    - Memory integration for conversation state
    - Tool integration for LMM completion stages
    """

    noop: ClassVar[Self]  # defined after the class
    """
    `noop` stage is a `Stage` which does nothing.
    """

    @overload
    @classmethod
    def predefined(
        cls,
        element: LMMContextElement,
        /,
        *elements: LMMContextElement,
        meta: Meta | MetaValues | None = None,
    ) -> Self: ...

    @overload
    @classmethod
    def predefined(
        cls,
        element: Prompt | Multimodal,
        /,
        *elements: Multimodal,
        meta: Meta | MetaValues | None = None,
    ) -> Self: ...

    @classmethod
    def predefined(
        cls,
        element: Prompt | LMMContextElement | Multimodal,
        /,
        *elements: LMMContextElement | Multimodal,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        """
        Creates a Stage that inserts predefined elements into the context.

        This Stage adds the specified elements to the LMM context and
        sets the result to the last provided completion value. It's useful for
        injecting static content into a processing pipeline.
        Note that context must always end with LMMCompletion as the result.

        Parameters
        ----------
        element : Prompt | LMMContextElement | Multimodal
            The first element to add to the context.
        *elements : LMMContextElement | Multimodal
            Additional elements to add to the context.
        meta: Meta | MetaValues | None = None
            Additional stage metadata including tags, description etc.

        Returns
        -------
        Self
            A Stage instance that adds predefined elements to the context.

        Examples
        --------
        >>> stage = Stage.predefined(
        ...     "What is 2+2?",
        ...     "The answer is 4."
        ... )
        """

        context: list[LMMContextElement]
        match element:
            case context_element if isinstance(context_element, LMMContextElement):
                context = [context_element]

            case Prompt() as prompt:
                context = [*prompt.content]

            case content:
                context = [LMMInput.of(MultimodalContent.of(content))]

        for idx, value in enumerate(elements):
            if idx % 2 == 0:
                if isinstance(value, LMMContextElement):
                    context.append(value)

                else:
                    context.append(LMMInput.of(MultimodalContent.of(value)))

            elif isinstance(value, LMMContextElement):
                context.append(value)

            else:
                context.append(LMMCompletion.of(MultimodalContent.of(value)))

        context_extension: LMMContext = tuple(context)
        assert isinstance(context_extension[-1], LMMCompletion)  # nosec: B101
        completion_result: MultimodalContent = context_extension[-1].content

        async def stage(
            *,
            state: StageState,
        ) -> StageState:
            with ctx.scope("stage.predefined"):
                return state.updated(
                    context=(*state.context, *context_extension),
                    result=completion_result,
                )

        return cls(stage, meta=Meta.of(meta))

    @classmethod
    def memory_recall(
        cls,
        memory: StageMemory,
        /,
        *,
        handling: Literal["replace", "merge"] = "replace",
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        """
        Creates a Stage that recalls context from memory.

        This Stage retrieves context from the provided memory and either replaces
        the current context or extends it, depending on the specified mode.

        Parameters
        ----------
        memory : Memory[LMMContext, Any]
            The memory instance from which to recall the LMM context.
        handling : Literal["replace", "merge"]
            Determines how the recalled context is used:
            - "replace": Completely replaces the current context with the recalled one.
            - "merge": Merges the recalled context with the current one.
            Default is "replace".
        meta: Meta | MetaValues | None = None
            Additional stage metadata including tags, description etc.

        Returns
        -------
        Self
            A new Stage instance that recalls context from memory.
        """

        match handling:
            case "replace":

                async def stage(
                    *,
                    state: StageState,
                ) -> StageState:
                    with ctx.scope("stage.memory_recall"):
                        return await memory.recall()

            case "merge":

                async def stage(
                    *,
                    state: StageState,
                ) -> StageState:
                    with ctx.scope("stage.memory_recall"):
                        return state.merged(await memory.recall())

        return cls(stage, meta=Meta.of(meta))

    @classmethod
    def memory_remember(
        cls,
        memory: StageMemory,
        /,
        *,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        """
        Creates a Stage that stores the current context in memory.

        This Stage saves the current LMM context to the provided memory instance
        without modifying the context or result.

        Parameters
        ----------
        memory : Memory[Any, LMMContext]
            The memory instance in which to store the current LMM context.
        meta: Meta | MetaValues | None = None
            Additional stage metadata including tags, description etc.

        Returns
        -------
        Self
            A new Stage instance that remembers context in memory.
        """

        async def stage(
            *,
            state: StageState,
        ) -> StageState:
            with ctx.scope("stage.memory_remember"):
                await memory.remember(state)
                return state

        return cls(stage, meta=Meta.of(meta))

    @classmethod
    def completion(
        cls,
        input: Prompt | Multimodal,  # noqa: A002
        /,
        *,
        instruction: Instruction | str | None = None,
        tools: Toolbox | Iterable[Tool] | None = None,
        output: LMMOutputSelection = "auto",
        meta: Meta | MetaValues | None = None,
        **extra: Any,
    ) -> Self:
        """
        Creates a Stage that generates a completion using an LMM.

        This Stage uses the current LMM context along with the provided input to
        generate a new completion. Both the input and the generated completion are
        added to the context, and the completion becomes the new result value.

        The Stage can utilize tools when specified, which may be invoked by the LMM
        during completion generation. Tool calls and their results are automatically
        managed and included in the context.

        Parameters
        ----------
        input : Prompt | Multimodal
            The input content to provide to the LMM, either as a Prompt or Multimodal.
        instruction : Instruction | str | None
            Optional instruction or guidance for the LMM.
        tools : Toolbox | Iterable[Tool] | None
            Optional tools that the LMM can use during completion generation.
        output : LMMOutputSelection
            Controls the modality/type of the generated completion, default is "auto".
        meta: Meta | MetaValues | None = None
            Additional stage metadata including tags, description etc.
        **extra : Any
            Additional parameters to pass to the LMM invocation.

        Returns
        -------
        Self
            A new Stage instance that generates a completion using the LMM.

        Raises
        ------
        RuntimeError
            If the LMM exceeds the limit of recursive tool calls.

        Examples
        --------
        >>> stage = Stage.completion(
        ...     "Calculate 25% of 80",
        ...     tools=[calculator]
        ... )
        """
        context_extension: LMMContext
        match input:
            case Prompt() as prompt:
                context_extension = prompt.content

            case input_content:
                context_extension = (LMMInput.of(MultimodalContent.of(input_content)),)

        toolbox: Toolbox = Toolbox.of(tools)

        async def stage(
            *,
            state: StageState,
        ) -> StageState:
            with ctx.scope("stage.completion"):
                context, result = await _lmm_completion(
                    instruction=instruction,
                    context=(*state.context, *context_extension),
                    toolbox=toolbox,
                    output=output,
                    **extra,
                )
                return state.updated(
                    context=context,
                    result=result,
                )

        return cls(stage, meta=Meta.of(meta))

    @classmethod
    def prompting_completion(
        cls,
        input: Callable[[], Coroutine[None, None, Multimodal]],  # noqa: A002
        /,
        *,
        instruction: Instruction | str | None = None,
        tools: Toolbox | Iterable[Tool] | None = None,
        output: LMMOutputSelection = "auto",
        meta: Meta | MetaValues | None = None,
        **extra: Any,
    ) -> Self:
        """
        Creates a Stage that generates a completion using an LMM after prompting for input.

        This Stage awaits the result of an asynchronous function to obtain input content,
        then uses this input along with the current LMM context to generate a completion.
        The obtained input and the generated completion are added to the context, and
        the completion becomes the new result value.

        Similar to the `completion` method, this Stage can utilize tools that may be
        invoked by the LMM during completion generation, with tool calls and their
        results managed automatically.

        Parameters
        ----------
        input : Callable[[], Coroutine[None, None, Multimodal]]
            An async function that returns the input content to provide to the LMM.
            This function will be awaited during stage execution.
        instruction : Instruction | str | None
            Optional instruction or guidance for the LMM.
        tools : Toolbox | Iterable[Tool] | None
            Optional tools that the LMM can use during completion generation.
        output : LMMOutputSelection
            Controls the modality/type of the generated completion, default is "auto".
        meta: Meta | MetaValues | None = None
            Additional stage metadata including tags, description etc.
        **extra : Any
            Additional parameters to pass to the LMM invocation.

        Returns
        -------
        Self
            A new Stage instance that generates a completion after prompting for input.

        Raises
        ------
        RuntimeError
            If the LMM exceeds the limit of recursive tool calls.

        Examples
        --------
        >>> async def get_user_input():
        ...     # Implementation that gets input from user
        ...     return "What is the capital of Poland?"
        ...
        >>> stage = Stage.prompting_completion(
        ...     get_user_input,
        ...     tools=[geography_tool]
        ... )
        """
        toolbox: Toolbox = Toolbox.of(tools)

        async def stage(
            *,
            state: StageState,
        ) -> StageState:
            with ctx.scope("stage.completion.prompting"):
                context, result = await _lmm_completion(
                    instruction=instruction,
                    context=(
                        *state.context,
                        LMMInput.of(MultimodalContent.of(await input())),
                    ),
                    toolbox=toolbox,
                    output=output,
                    **extra,
                )
                return state.updated(
                    context=context,
                    result=result,
                )

        return cls(stage, meta=Meta.of(meta))

    @classmethod
    def loopback_completion(
        cls,
        *,
        instruction: Instruction | str | None = None,
        tools: Toolbox | Iterable[Tool] | None = None,
        output: LMMOutputSelection = "auto",
        meta: Meta | MetaValues | None = None,
        **extra: Any,
    ) -> Self:
        """
        Creates a Stage that takes the last completion and uses it as the new input.

        This Stage takes the last LMMContext completion, replaces the last input with it,
        and executes a new completion using that input. It's useful for iterative refinement
        of outputs where each completion builds upon the previous one.

        Parameters
        ----------
        instruction : Instruction | str | None
            Optional instruction or guidance for the LMM.
        tools : Toolbox | Iterable[Tool] | None
            Optional tools that the LMM can use during completion generation.
        output : LMMOutputSelection
            Controls the modality/type of the generated completion, default is "auto".
        meta: Meta | MetaValues | None = None
            Additional stage metadata including tags, description etc.
        **extra : Any
            Additional parameters to pass to the LMM invocation.

        Returns
        -------
        Self
            A new Stage instance that performs the loopback completion.

        Raises
        ------
        RuntimeError
            If the LMM exceeds the limit of recursive tool calls.
        """
        toolbox: Toolbox = Toolbox.of(tools)

        async def stage(
            *,
            state: StageState,
        ) -> StageState:
            with ctx.scope("stage.completion.loopback"):
                if not state.context or not isinstance(state.context[-1], LMMCompletion):
                    ctx.log_warning("loopback_completion has been skipped due to invalid context")
                    return state

                # Find the index of the last LMMInput in the context
                last_input_idx: int = -1
                for idx, element in enumerate(reversed(state.context)):
                    if isinstance(element, LMMInput):
                        last_input_idx = len(state.context) - idx - 1
                        break

                else:
                    ctx.log_warning("loopback_completion could not find an LMMInput in the context")
                    return state

                context, result = await _lmm_completion(
                    instruction=instruction,
                    context=(
                        *state.context[:last_input_idx],
                        LMMInput.of(state.context[-1].content),
                    ),
                    toolbox=toolbox,
                    output=output,
                    **extra,
                )
                return state.updated(
                    context=context,
                    result=result,
                )

        return cls(stage, meta=Meta.of(meta))

    @classmethod
    def transform_result(
        cls,
        transformation: StageResultTransforming,
        /,
        *,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        """
        Creates a Stage that transforms the current result without altering the context.

        This Stage applies a transformation function to the current result to produce a new
        result value, while keeping the LMMContext unchanged.

        Parameters
        ----------
        transformation : StageResultTransforming
            Function that takes the current result and returns a transformed result.
        meta: Meta | MetaValues | None = None
            Additional stage metadata including tags, description etc.

        Returns
        -------
        Self
            A new Stage instance that transforms the result.
        """

        async def stage(
            *,
            state: StageState,
        ) -> StageState:
            with ctx.scope("stage.transform.result"):
                return state.updated(result=await transformation(state.result))

        return cls(stage, meta=Meta.of(meta))

    @classmethod
    def transform_context(
        cls,
        transformation: StageContextTransforming,
        /,
        *,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        """
        Creates a Stage that transforms the current context without altering the result.

        This Stage applies a transformation function to the current LMMContext to produce a new
        context, while keeping the result unchanged.

        Parameters
        ----------
        transformation : StageContextTransforming
            Function that takes the current context and returns a transformed context.
        meta: Meta | MetaValues | None = None
            Additional stage metadata including tags, description etc.

        Returns
        -------
        Self
            A new Stage instance that transforms the context.
        """

        async def stage(
            *,
            state: StageState,
        ) -> StageState:
            with ctx.scope("stage.transform.context"):
                return state.updated(context=await transformation(state.context))

        return cls(stage, meta=Meta.of(meta))

    @classmethod
    def trim_context(
        cls,
        *,
        limit: slice | int | None = None,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        """
        Creates a Stage that trims the current context to a specified limit.

        This Stage reduces the size of the LMMContext by applying a slice or index limit,
        or by clearing it entirely if no limit is specified.

        Parameters
        ----------
        limit : slice | int | None
            Specifies how to trim the context:
            - None: Clear the context completely
            - int: Keep only the first N elements
            - slice: Apply a slice operation to the context
        meta: Meta | MetaValues | None = None
            Additional stage metadata including tags, description etc.

        Returns
        -------
        Self
            A new Stage instance that trims the context.
        """
        match limit:
            case None:

                async def stage(
                    *,
                    state: StageState,
                ) -> StageState:
                    return state.updated(context=())

            case int() as index:

                async def stage(
                    *,
                    state: StageState,
                ) -> StageState:
                    return state.updated(context=state.context[:index])

            case slice() as index_slice:

                async def stage(
                    *,
                    state: StageState,
                ) -> StageState:
                    return state.updated(context=state.context[index_slice])

        return cls(stage, meta=Meta.of(meta))

    @classmethod
    def strip_context_tools(
        cls,
        *,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        """
        Creates a Stage that trims the current context by removing tool calls.

        Parameters
        ----------
        meta: Meta | MetaValues | None = None
            Additional stage metadata including tags, description etc.

        Returns
        -------
        Self
            A new Stage instance that removes tool-related elements from the context.
        """

        async def stage(
            *,
            state: StageState,
        ) -> StageState:
            return state.updated(
                context=tuple(
                    element
                    for element in state.context
                    if not isinstance(element, LMMToolRequests | LMMToolResponses)
                ),
            )

        return cls(stage, meta=Meta.of(meta))

    @classmethod
    def loop(
        cls,
        stage: Self,
        /,
        *stages: Self,
        condition: StageLoopConditioning,
        mode: Literal[
            "pre_check",
            "post_check",
        ] = "post_check",
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        """
        Creates a Stage that executes another Stage in a loop while a condition is met.

        This Stage repeatedly executes the provided Stage as long as the condition
        function returns True when applied to the current context and result.

        Parameters
        ----------
        stage : Stage
            The Stage to execute repeatedly.
        condition : StageLoopConditioning
            A function that determines whether to continue looping. It should be
            an async function that accepts the current state and iteration number
            and returns a boolean.
        mode: Literal["pre_check", "post_check"]
            A loop mode determining the first condition check behavior.
            "pre_check" will check the condition before executing stage
            "post_check" will check the condition after executing stage
        meta: Meta | MetaValues | None = None
            Additional stage metadata including tags, description etc.

        Returns
        -------
        Stage
            A new Stage instance that implements the looping behavior.
        """
        stage_executions: Sequence[StageExecution] = (
            stage,
            *(stage for stage in stages),
        )

        match mode:
            case "pre_check":

                async def stage_loop(
                    *,
                    state: StageState,
                ) -> StageState:
                    with ctx.scope("stage.loop"):
                        current_state: StageState = state
                        iteration: int = 0

                        while await condition(state=current_state, iteration=iteration):
                            with ctx.scope("stage.loop.iteration"):
                                for execution in stage_executions:
                                    current_state = await execution(state=current_state)

                                iteration += 1

                        return current_state

            case "post_check":

                async def stage_loop(
                    *,
                    state: StageState,
                ) -> StageState:
                    with ctx.scope("stage.loop"):
                        current_state: StageState = state
                        iteration: int = 0

                        while True:
                            with ctx.scope("stage.loop.iteration"):
                                for stage_execution in stage_executions:
                                    current_state = await stage_execution(state=current_state)

                                if not await condition(state=current_state, iteration=iteration):
                                    return current_state

                                iteration += 1

        return cls(stage_loop, meta=Meta.of(meta))

    @classmethod
    def sequence(
        cls,
        stage: Self,
        /,
        *stages: Self,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        """
        Creates a Stage that executes multiple Stages in sequence.

        This Stage chains the execution of the provided Stages, where each Stage
        receives the context and result produced by the previous Stage.

        Parameters
        ----------
        stage : Stage
            The first Stage in the sequence.
        *stages : Stage
            Additional Stages to execute in order after the first Stage.
        meta: Meta | MetaValues | None = None
            Additional stage metadata including tags, description etc.

        Returns
        -------
        Stage
            A new Stage instance that executes the provided Stages sequentially.
        """
        stage_executions: Sequence[StageExecution] = tuple(stage for stage in (stage, *stages))

        async def stage_sequence(
            *,
            state: StageState,
        ) -> StageState:
            current_state: StageState = state

            for execution in stage_executions:
                current_state = await execution(state=current_state)

            return current_state

        return cls(stage_sequence, meta=Meta.of(meta))

    @classmethod
    def concurrent(
        cls,
        stage: Self,
        /,
        *stages: Self,
        merge: StageMerging,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        """
        Creates a Stage that executes multiple Stages concurrently.

        This Stage runs all provided Stages in parallel using the same initial
        context and result, then combines their results using the specified
        merge function.

        Parameters
        ----------
        stage : Stage
            The first Stage to execute concurrently.
        *stages : Stage
            Additional Stages to execute concurrently with the first Stage.
        merge : StageMerging
            A function that merges the results from all concurrent Stage executions.
            It should be an async function that accepts a sequence of (context, result)
            tuples and returns a single (context, result) tuple.
        meta: Meta | MetaValues | None = None
            Additional stage metadata including tags, description etc.

        Returns
        -------
        Stage
            A new Stage instance that executes Stages concurrently and merges results.
        """
        stage_executions: Sequence[StageExecution] = tuple(stage for stage in (stage, *stages))

        async def concurrent_stage(
            *,
            state: StageState,
        ) -> StageState:
            with ctx.scope("stage.concurrent"):
                return await merge(
                    branches=cast(  # we are converting errors witin __call__
                        Sequence[StageState | StageException],
                        await gather(
                            *[execution(state=state) for execution in stage_executions],
                            return_exceptions=True,
                        ),
                    )
                )

        return cls(concurrent_stage, meta=Meta.of(meta))

    @classmethod
    def router(
        cls,
        stage: Self,
        /,
        *stages: Self,
        routing: StageRouting | None = None,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        """
        Creates a Stage that routes execution to one of several stages based on a routing function.

        This Stage uses a routing function to select which of the provided stages to execute
        based on the current context and result. Each stage is identified by its metadata name
        (required) and description (required). Stage names are normalized to lowercase for matching.

        By default, uses LLM-based routing that analyzes the context and result to intelligently
        select the most appropriate stage based on stage descriptions.

        Parameters
        ----------
        stage : Stage
            The first stage option for routing. Must have name and description in metadata.
        *stages : Stage
            Additional stage options for routing. Each must have name and description in metadata.
        routing : StageRouting, optional
            A function that selects which stage to execute. Defaults to `_lmm_routing` which
            uses LLM analysis. The function receives the current context, result, and available
            options, and returns the key of the selected stage.
        meta: Meta | MetaValues | None = None
            Additional stage metadata including tags, description etc.

        Returns
        -------
        Self
            A new Stage instance that routes execution to the selected stage.

        Raises
        ------
        AssertionError
            If any stage is missing required name or description metadata.
        ValueError
            If the routing function returns a selection that doesn't match any available stage.

        Examples
        --------
        >>> # Using default LLM-based routing
        >>> router = Stage.router(
        ...     normal_stage.with_meta(
        ...         name="process_data",
        ...         description="Process and analyze the input data"
        ...     ),
        ...     error_stage.with_meta(
        ...         name="handle_error",
        ...         description="Handle errors and provide error responses"
        ...     )
        ... )
        >>>
        >>> # Using custom routing function
        >>> async def custom_routing(
        ...     context: LMMContext,
        ...     result: MultimodalContent,
        ...     options: Mapping[str, Meta],
        ... ):
        ...     if "error" in result.as_string():
        ...         return "handle_error"
        ...     return "process_data"
        ...
        >>> router = Stage.router(
        ...     normal_stage.with_meta(name="process_data", description="..."),
        ...     error_stage.with_meta(name="handle_error", description="..."),
        ...     routing=custom_routing
        ... )
        """
        if not stages:  # there is no selection when there is only one option
            ctx.log_debug("Stage.router prepared with single option, skipping routing...")
            return stage

        routes: MutableMapping[str, StageExecution] = {}
        options: MutableMapping[str, Meta] = {}
        for route in (stage, *stages):
            assert route.meta.name, "Stage names are required for routing"  # nosec: B101
            assert route.meta.description, "Stage descriptions are required for routing"  # nosec: B101
            key: str = (route.meta.name or uuid4().hex).strip().lower()
            assert "\n" not in key, "Stage names can't have newlines"  # nosec: B101
            routes[key] = route._execution
            options[key] = route.meta

        routing = routing if routing is not None else _lmm_routing

        async def router_stage(
            *,
            state: StageState,
        ) -> StageState:
            with ctx.scope("stage.router"):
                selection: str = await routing(
                    state=state,
                    options=options,
                )

            # out of scope - executing selected
            if selected := routes.get(selection):
                return await selected(state=state)

            else:
                raise RuntimeError(
                    "Stage.router selection invalid"
                    f" - missing selection ({selection}) in available options"
                )

        return cls(router_stage, meta=Meta.of(meta))

    __slots__ = (
        "_execution",
        "meta",
    )

    def __init__(
        self,
        execution: StageExecution,
        /,
        meta: Meta,
    ) -> None:
        assert not isinstance(execution, Stage)  # nosec: B101
        assert isinstance(execution, StageExecution)  # nosec: B101
        self._execution: StageExecution
        object.__setattr__(
            self,
            "_execution",
            execution,
        )
        self.meta: Meta
        object.__setattr__(
            self,
            "meta",
            meta,
        )

    def with_meta(
        self,
        **values: MetaValue,
    ) -> Self:
        """
        Creates a copy of this Stage with updated metadata.

        This method allows you to add or update metadata fields for the Stage,
        such as name, description, tags, and other custom metadata values.
        The metadata is particularly important for routing stages where name
        and description are required.

        Parameters
        ----------
        **values : MetaValue
            Metadata values to add or update. Common values include:
            - name: str - A unique identifier for the stage
            - description: str - A description of what the stage does
            - tags: Sequence[str] - Tags for categorizing the stage
            - Any other custom metadata fields

        Returns
        -------
        Self
            A new Stage instance with the updated metadata.

        Examples
        --------
        >>> stage = Stage.completion("Calculate 2+2")
        >>> named_stage = stage.with_meta(
        ...     name="calculator",
        ...     description="Performs basic arithmetic calculations",
        ...     tags=["math", "arithmetic"]
        ... )
        >>>
        >>> # Required for routing stages
        >>> router = Stage.router(
        ...     stage1.with_meta(name="option1", description="First option"),
        ...     stage2.with_meta(name="option2", description="Second option"),
        ...     routing=my_routing_function
        ... )
        """
        return self.__class__(
            self._execution,
            meta=self.meta.merged_with(values),
        )

    @overload
    def with_ctx(
        self,
        /,
        *,
        disposables: Disposables | Iterable[Disposable],
    ) -> Self: ...

    @overload
    def with_ctx(
        self,
        state_context: State,
        /,
        *state: State,
    ) -> Self: ...

    def with_ctx(
        self,
        state: State | None = None,
        /,
        *states: State,
        disposables: Disposables | Iterable[Disposable] | None = None,
    ) -> Self:
        """
        Creates a copy of this Stage with the specified state context applied.

        The returned Stage will execute within the provided state context,
        which affects what contextual state is available.

        Note that Processing state will be replaced with contextual.

        Parameters
        ----------
        state_context : State | None
            `State` object to include in the state context update.

        *state : State
            Optional additional `State` objects to include in the state context.
            Only applicable when `state_context` is a `State`.

        disposables: Disposables | Iterable[Disposable] | None
            Optional Disposables which will be used for execution of this stage.
            State produced by disposables will be used within the context state.

        Returns
        -------
        Stage
            A new Stage instance that executes within the specified context.
        """
        execution: StageExecution = self._execution

        resolved_disposables: Disposables | None
        match disposables:
            case None:
                resolved_disposables = None

            case Disposables() as disposables:
                resolved_disposables = disposables

            case iterable:
                resolved_disposables = Disposables(*iterable)

        match (state, resolved_disposables):
            case (None, None):
                assert not states  # nosec: B101
                return self  # nothing to change...

            case (None, ctx_disposables):
                assert not states  # nosec: B101

                async def stage(
                    *,
                    state: StageState,
                ) -> StageState:
                    async with ctx_disposables as features:
                        with ctx.updated(*features):
                            return await execution(state=state)

            case (ctx_state, None):

                async def stage(
                    *,
                    state: StageState,
                ) -> StageState:
                    with ctx.updated(ctx_state, *states, ctx.state(Processing)):
                        return await execution(state=state)

            case (ctx_state, ctx_disposables):

                async def stage(
                    *,
                    state: StageState,
                ) -> StageState:
                    async with ctx_disposables as features:
                        with ctx.updated(
                            ctx_state,
                            *states,
                            *features,
                            # preserve current Processing state by replacing it
                            ctx.state(Processing),
                        ):
                            return await execution(state=state)

        return self.__class__(stage, meta=self.meta)

    @overload
    def cached[Key: Hashable](
        self,
        *,
        limit: int | None = None,
        expiration: float | None = None,
        make_key: MakeCacheKey[Key] | None = None,
    ) -> Self: ...

    @overload
    def cached[Key](
        self,
        *,
        make_key: MakeCacheKey[Key],
        read: ReadCache[Key],
        write: WriteCache[Key],
    ) -> Self: ...

    def cached[Key](
        self,
        *,
        limit: int | None = None,
        expiration: float | None = None,
        make_key: MakeCacheKey[Key] | None = None,
        read: ReadCache[Key] | None = None,
        write: WriteCache[Key] | None = None,
    ) -> Self:
        """
        Creates a copy of this Stage with caching enabled.

        The returned Stage will provide the same context and result as the last execution
        if cache keys match.
        Note that the result will be the exact same context and result each time regarless
        of cached stage actual execution assuming no side effects or processing state changes
        are produced within cached stage.

        Parameters
        ----------
        limit : int | None
            Maximum number of entries to store in the cache. If None, cache size is
            one. Ignored when using custom read and write.
        expiration : float | None
            Time in seconds after which cache entries expire. If None, entries never
            expire. Ignored when using custom read and write.
        make_key : MakeCacheKey[Key] | None
            Custom function to generate cache keys from execution inputs. The function
            should accept context and result parameters and return a hashable value
            uniquely identifying the inputs.
        read : ReadCache[Key] | None
            Custom function to retrieve cached values. Must be provided together with
            `write` and `make_key`.
        write : WriteCache[Key] | None
            Custom function to store values in cache. Must be provided together with
            `read` and `make_key`.

        Returns
        -------
        Self
            A new Stage instance with caching enabled.
        """

        if read is not None and write is not None and make_key is not None:
            return self.__class__(
                cast(
                    StageExecution,
                    cache(
                        make_key=make_key,
                        read=read,
                        write=write,
                    )(self._execution),
                ),
                meta=self.meta,
            )

        else:
            assert read is None and write is None and make_key is None  # nosec: B101
            return self.__class__(
                cache(
                    limit=limit,
                    expiration=expiration,
                    make_key=make_key,
                )(self._execution),
                meta=self.meta,
            )

    def with_retry(
        self,
        *,
        limit: int = 1,
        delay: Callable[[int, Exception], float] | float | None = None,
        catching: set[type[Exception]] | tuple[type[Exception], ...] | type[Exception] = Exception,
    ) -> Self:
        """
        Creates a copy of this Stage with retry behavior added.

        The returned Stage will re-execute its processing function if it raises
        any of the exceptions specified in `catching`, up to `limit` times.
        Retries can be delayed using a fixed `delay` in seconds or a dynamic
        `delay` function that calculates the delay based on the retry attempt and exception.

        Parameters
        ----------
        limit : int
            Maximum number of retry attempts. Defaults to 1.
        delay : Callable[[int, Exception], float] | float | None
            Delay between retries in seconds. Can be a fixed float, a function
            that accepts the retry attempt number (starting from 0) and the raised
            exception and returns the delay, or None for no delay. Defaults to None.
        catching : set[type[Exception]] | tuple[type[Exception], ...] | type[Exception]
            Exception types to catch and retry on. Defaults to catching all Exceptions.

        Returns
        -------
        Self
            A new Stage instance with retry behavior.
        """

        return self.__class__(
            retry(
                limit=limit,
                delay=delay,
                catching=catching,
            )(self._execution),
            meta=self.meta,
        )

    def with_fallback(
        self,
        fallback: Self,
        *,
        catching: set[type[Exception]] | tuple[type[Exception], ...] | type[Exception] = Exception,
    ) -> Self:
        """
        Creates a copy of this Stage with fallback behavior added.

        The returned Stage will execute fallback stage if it raises
        any of the exceptions specified in `catching`.

        Parameters
        ----------
        fallback : Stage
            Stage to be executed as fallback.
        catching : set[type[Exception]] | tuple[type[Exception], ...] | type[Exception]
            Exception types to catch and fallback on. Defaults to catching all Exceptions.

        Returns
        -------
        Self
            A new Stage instance with fallback behavior.
        """

        execution: StageExecution = self._execution
        fallback_execution: StageExecution = fallback._execution
        exceptions: Collection[type[Exception]] = (
            catching if isinstance(catching, set | tuple) else {catching}
        )

        async def stage(
            *,
            state: StageState,
        ) -> StageState:
            try:
                return await execution(state=state)

            except Exception as exc:
                with ctx.scope("stage.fallback"):
                    if exc in exceptions:
                        return await fallback_execution(state=state)

                    else:
                        raise exc

        return self.__class__(stage, meta=self.meta)

    def with_volatile_context(self) -> Self:
        """
        Creates a copy of this Stage that discards context changes.

        The returned Stage will execute normally but will revert any changes
        made to the LMM context when completed, keeping only the result changes.

        Returns
        -------
        Stage
            A new Stage instance that discards context changes.
        """
        execution: StageExecution = self._execution

        async def stage(
            *,
            state: StageState,
        ) -> StageState:
            processed_state: StageState = await execution(state=state)
            return processed_state.updated(context=state.context)

        return self.__class__(stage, meta=self.meta)

    def with_volatile_tools_context(self) -> Self:
        """
        Creates a copy of this Stage that discards tool calls added by this stage.

        The returned Stage will execute normally, but upon completion, it will
        remove any LMMToolRequests and LMMToolResponses elements that were added
        to the context during the execution of this specific stage. When produced
        and a previous contexts share a common prefix with tools usage it won't be
        removed from the result context.

        Returns
        -------
        Stage
            A new Stage instance that makes tool-related context elements volatile.
        """
        execution: StageExecution = self._execution

        async def stage(
            *,
            state: StageState,
        ) -> StageState:
            processed_state: StageState = await execution(state=state)

            common_prefix: LMMContext = tuple(
                current
                for previous, current in zip(
                    state.context,
                    processed_state.context,
                    strict=False,
                )
                if current == previous
            )
            striped_suffix: LMMContext = tuple(
                element
                for element in processed_state.context[len(common_prefix) :]
                if not isinstance(element, LMMToolRequests | LMMToolResponses)
            )

            return state.updated(
                context=(*common_prefix, *striped_suffix),
                result=processed_state.result,
            )

        return self.__class__(stage, meta=self.meta)

    def when(
        self,
        condition: StageConditioning | bool,
        /,
        *,
        alternative: Self | None = None,
    ) -> Self:
        """
        Creates a Stage that conditionally executes based on a predicate.

        This method returns a new Stage that will only execute the current Stage
        if the specified condition is met. If the condition is not met and an
        alternative Stage is provided, that alternative will be executed instead.

        Parameters
        ----------
        condition : StageCondition | bool
            A boolean value or function that determines whether this Stage executes.
            If a function is provided, it will be called with the current context and result.
        alternative : Stage | None
            Optional Stage to execute when the condition is not met.

        Returns
        -------
        Stage
            A new Stage instance that applies conditional execution logic.

        Examples
        --------
        >>> # Execute stage only when result contains specific text
        >>> async def has_error(context: LMMContext, result: MultimodalContent):
        ...     return "error" in result.as_string()
        >>>
        >>> error_handling = my_stage.when(has_error)
        >>>
        >>> # Use a boolean condition with an alternative
        >>> primary_stage.when(is_premium_user, alternative=basic_stage)
        """
        execution: StageExecution = self._execution
        alternative_execution: StageExecution | None = (
            alternative._execution if alternative else None
        )

        match condition:
            case bool() as value:

                async def stage(
                    *,
                    state: StageState,
                ) -> StageState:
                    if value:
                        return await execution(state=state)

                    elif alternative_execution:
                        return await alternative_execution(state=state)

                    else:
                        return state

            case function:

                async def stage(
                    *,
                    state: StageState,
                ) -> StageState:
                    if await function(state=state):
                        return await execution(state=state)

                    elif alternative_execution:
                        return await alternative_execution(state=state)

                    else:
                        return state

        return self.__class__(stage, meta=self.meta)

    def ignore_result(self) -> Self:
        """
        Creates a copy of this Stage that ignores its produced result.

        The returned Stage will execute normally but will retain the original
        result value rather than replacing it with the result of execution.

        Returns
        -------
        Stage
            A new Stage instance that ignores its produced result.
        """
        execution: StageExecution = self._execution

        async def stage(
            *,
            state: StageState,
        ) -> StageState:
            processed_state: StageState = await execution(state=state)
            return processed_state.updated(result=state.result)  # preserve current result

        return self.__class__(stage, meta=self.meta)

    def extend_result(self) -> Self:
        """
        Creates a copy of this Stage that extends the current result.

        The returned Stage will execute normally but will append its result
        to the current result rather than replacing it.

        Returns
        -------
        Stage
            A new Stage instance that extends rather than replaces the result.
        """
        execution: StageExecution = self._execution

        async def stage(
            *,
            state: StageState,
        ) -> StageState:
            processed_state: StageState = await execution(state=state)
            return processed_state.updated(
                result=state.result.appending(processed_state.result),
            )

        return self.__class__(stage, meta=self.meta)

    def __setattr__(
        self,
        __name: str,
        __value: Any,
    ) -> None:
        raise RuntimeError("Stage is frozen and can't be modified")

    def __delattr__(
        self,
        __name: str,
    ) -> None:
        raise RuntimeError("Stage is frozen and can't be modified")

    async def __call__(
        self,
        *,
        state: StageState,
    ) -> StageState:
        try:
            return await self._execution(state=state)

        except StageException as exc:
            raise exc

        except BaseException as exc:
            raise StageException(
                f"Stage execution interrupted: {exc}",
                state=state,
            ) from exc

    async def execute(
        self,
        *state: DataModel | State,
        context: LMMContext | None = None,
        result: MultimodalContent | None = None,
    ) -> MultimodalContent:
        """
        Executes this Stage with the provided initial values.

        This method runs the Stage's processing function with the specified
        initial context, result, and state, and returns the produced result.

        Parameters
        ----------
        *state : DataModel | State
            The initial state models to use.
        context : LMMContext | None
            The initial LMM context to use, or None to use an empty context.
        result : MultimodalContent | None
            The initial result value to use, or None to use an empty result.

        Returns
        -------
        MultimodalContent
            The result produced by executing the Stage.
        """
        with ctx.scope("stage.execution"):
            initial_context: LMMContext
            match context:
                case None:
                    initial_context = ()

                case context:
                    assert not context or isinstance(context[-1], LMMCompletion)  # nosec: B101
                    initial_context = context

            result_state: StageState = await self(
                state=StageState.of(
                    *state,
                    context=initial_context,
                    result=result if result is not None else MultimodalContent.empty,
                )
            )

            return result_state.result


async def _lmm_completion(
    *,
    instruction: Instruction | str | None,
    context: LMMContext,
    toolbox: Toolbox,
    output: LMMOutputSelection,
    **extra: Any,
) -> tuple[LMMContext, MultimodalContent]:
    current_context: LMMContext = context
    repetition_level: int = 0
    while True:
        match await LMM.completion(
            instruction=instruction,
            context=current_context,
            output=output,
            tools=toolbox.available_tools(repetition_level=repetition_level),
            tool_selection=toolbox.tool_selection(repetition_level=repetition_level),
            **extra,
        ):
            case LMMCompletion() as completion:
                return (
                    (*current_context, completion),
                    completion.content,
                )

            case LMMToolRequests() as tool_requests:
                tool_responses: LMMToolResponses = await toolbox.respond_all(tool_requests)

                if direct_results := [
                    response.content
                    for response in tool_responses.responses
                    if response.handling == "direct_result"
                ]:
                    direct_content: MultimodalContent = MultimodalContent.of(*direct_results)

                    return (
                        (
                            *current_context,
                            LMMCompletion.of(direct_content),
                        ),
                        direct_content,
                    )

                else:
                    current_context = (
                        *current_context,
                        tool_requests,
                        tool_responses,
                    )

        repetition_level += 1  # continue tracking repetetive tool calls

    raise RuntimeError("LMM exceeded limit of recursive calls")


async def _lmm_routing(
    *,
    state: StageState,
    options: Mapping[str, Meta],
) -> str:
    options_text = "\n".join(
        [f"- {key}: {meta.description or 'No description'}" for key, meta in options.items()]
    )

    instruction: str = (
        "Based on the conversation context and current result,"  # nosec: B608 - false positive
        " select the most appropriate option from the following:"
        f"\n\n{options_text}"
        "\n\nRespond with with the exact option name within SELECTION xml tag"
        f" like (e.g., '<SELECTION>{next(iter(options.keys()))}</SELECTION>'"
    )

    # Create routing context with the current result as input
    routing_context: LMMContext = (
        *state.context,
        LMMInput.of(
            MultimodalContent.of(
                "<RESULT>",
                state.result,
                "</RESULT>",
            )
        ),
    )

    selection: str
    match await LMM.completion(
        instruction=instruction,
        context=routing_context,
        output="text",
    ):
        case LMMCompletion() as completion:
            if selection_tag := MultimodalTagElement.parse_first(
                "SELECTION",
                content=completion.content,
            ):
                selection = selection_tag.content.to_str().strip().lower()

            else:
                selection = "__INVALID_SELECTION__"

        case other:
            raise RuntimeError(f"LMM routing failed: unexpected result {type(other)}")

    if selection not in options:
        raise ValueError(
            f"LMM routing failed: invalid selection '{selection}'. "
            f"Available options: {list(options.keys())}"
        )

    return selection


@overload
def stage(
    execution: StageExecution,
    /,
) -> Stage: ...


@overload
def stage(
    *,
    meta: Meta | MetaValues | None = None,
) -> Callable[[StageExecution], Stage]: ...


def stage(
    execution: StageExecution | None = None,
    /,
    *,
    meta: Meta | MetaValues | None = None,
) -> Callable[[StageExecution], Stage] | Stage:
    def wrap(execution: StageExecution) -> Stage:
        return Stage(execution, meta=Meta.of(meta))

    if execution is None:
        return wrap

    else:
        return wrap(execution)


@stage
async def _noop(
    *,
    state: StageState,
) -> StageState:
    return state


Stage.noop = _noop



from pathlib import Path

from draive.mistral.api import MistralAPI
from draive.sentencepiece import sentencepiece_processor, sentencepiece_tokenizer
from draive.tokenization import Tokenization

__all__ = ("MistralTokenization",)


class MistralTokenization(MistralAPI):
    async def tokenizer(
        self,
        model_name: str,
        /,
    ) -> Tokenization:
        """
        Prepare tokenizer for selected Mistral model.
        """

        return Tokenization(
            text_tokenizing=sentencepiece_tokenizer(
                await sentencepiece_processor(model_path=_model_path(model_name))
            )
        )


def _model_path(
    model_name: str,
    /,
) -> str:
    model_file: str = _mapping.get(model_name, "mistral_v3_tokenizer.model")
    return str(Path(__file__).parent / "tokens" / model_file)


_mapping: dict[str, str] = {
    "open-mistral-7b": "mistral_v1_tokenizer.model",
    "open-mixtral-8x7b": "mistral_v1_tokenizer.model",
    "mistral-embed": "mistral_v1_tokenizer.model",
    "mistral-small": "mistral_v2_tokenizer.model",
    "mistral-large": "mistral_v2_tokenizer.model",
    "open-mixtral-8x22b": "mistral_v3_tokenizer.model",
    "codestral-22b": "mistral_v3_tokenizer.model",
}



from collections.abc import AsyncIterator, Iterable
from typing import Any, Literal, cast, final, overload

from haiway import State, ctx

from draive.conversation.completion.default import conversation_completion
from draive.conversation.completion.types import ConversationCompleting
from draive.conversation.types import (
    ConversationElement,
    ConversationMemory,
    ConversationMessage,
    ConversationStreamElement,
)
from draive.helpers import ConstantMemory
from draive.instructions import Instruction
from draive.multimodal import Multimodal
from draive.tools import Tool, Toolbox
from draive.utils import Memory

__all__ = ("Conversation",)


@final
class Conversation(State):
    @overload
    @classmethod
    async def completion(
        cls,
        *,
        instruction: Instruction | str | None = None,
        input: ConversationMessage | Multimodal,
        memory: ConversationMemory | Iterable[ConversationElement] | None = None,
        tools: Toolbox | Iterable[Tool] | None = None,
        stream: Literal[False] = False,
        **extra: Any,
    ) -> ConversationMessage: ...

    @overload
    @classmethod
    async def completion(
        cls,
        *,
        instruction: Instruction | str | None = None,
        input: ConversationMessage | Multimodal,
        memory: ConversationMemory | Iterable[ConversationElement] | None = None,
        tools: Toolbox | Iterable[Tool] | None = None,
        stream: Literal[True],
        **extra: Any,
    ) -> AsyncIterator[ConversationStreamElement]: ...

    @classmethod
    async def completion(
        cls,
        *,
        instruction: Instruction | str | None = None,
        input: ConversationMessage | Multimodal,  # noqa: A002
        memory: ConversationMemory | Iterable[ConversationElement] | None = None,
        tools: Toolbox | Iterable[Tool] | None = None,
        stream: bool = False,
        **extra: Any,
    ) -> AsyncIterator[ConversationStreamElement] | ConversationMessage:
        conversation: Conversation = ctx.state(cls)

        # prepare memory
        conversation_memory: ConversationMemory
        match memory if memory is not None else conversation.memory:
            case None:
                conversation_memory = ConstantMemory(recalled=())

            case Memory() as memory:
                conversation_memory = memory

            case memory_messages:
                conversation_memory = cast(
                    ConversationMemory,
                    ConstantMemory(recalled=tuple(memory_messages)),
                )

        if stream:
            return await conversation.completing(
                instruction=Instruction.of(instruction),
                input=input,
                memory=conversation_memory,
                toolbox=Toolbox.of(tools),
                stream=True,
            )

        else:
            return await conversation.completing(
                instruction=Instruction.of(instruction),
                input=input,
                memory=conversation_memory,
                toolbox=Toolbox.of(tools),
            )

    completing: ConversationCompleting = conversation_completion
    memory: ConversationMemory | None = None



from asyncio import gather
from collections.abc import Collection, Iterable, Mapping, Sequence
from typing import Any, Literal, Self, final, overload

from haiway import State, ctx

from draive.commons import META_EMPTY, Meta, MetaTags, MetaValues
from draive.lmm.types import (
    LMMException,
    LMMToolError,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolResponse,
    LMMToolResponseHandling,
    LMMToolResponses,
    LMMToolSpecification,
)
from draive.multimodal import MultimodalContent
from draive.tools.state import Tools
from draive.tools.types import Tool

__all__ = ("Toolbox",)


@final
class Toolbox(State):
    @classmethod
    def of(  # noqa: C901, PLR0912, PLR0915
        cls,
        tool_or_toolbox: Self | Tool | Iterable[Tool] | None = None,
        /,
        *tools: Tool,
        suggest: Tool | str | bool | None = None,
        repeated_calls_limit: int | None = None,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        tools_mapping: Mapping[str, Tool]
        suggest_call: Tool | bool
        calls_limit: int
        metadata: Meta
        match tool_or_toolbox:
            case None:
                assert suggest is None or suggest is False  # nosec: B101
                tools_mapping = {}
                suggest_call = False
                calls_limit = 0
                metadata = Meta.of(meta)

            case Toolbox() as toolbox:
                tools_mapping = {
                    **toolbox.tools,
                    **{tool.name: tool for tool in tools},
                }

                match suggest if suggest is not None else toolbox.suggest_call:
                    case bool() as suggestion:
                        suggest_call = suggestion

                    case str() as tool_name:
                        suggest_call = tools_mapping[tool_name]

                    case Tool() as tool:
                        assert tool.name in tools_mapping  # nosec: B101
                        suggest_call = tool

                calls_limit = (
                    repeated_calls_limit
                    if repeated_calls_limit is not None
                    else toolbox.repeated_calls_limit
                )
                metadata = (
                    toolbox.meta.merged_with(Meta.of(meta)) if meta is not None else toolbox.meta
                )

            case Tool() as tool:
                tools_mapping = {
                    **{tool.name: tool},
                    **{tool.name: tool for tool in tools},
                }

                match suggest:
                    case None:
                        suggest_call = False

                    case bool() as suggestion:
                        suggest_call = suggestion

                    case str() as tool_name:
                        suggest_call = tools_mapping[tool_name]

                    case Tool() as tool:
                        assert tool.name in tools_mapping  # nosec: B101
                        suggest_call = tool

                calls_limit = repeated_calls_limit or 3
                metadata = Meta.of(meta)

            case iterable_tools:
                tools_mapping = {
                    **{tool.name: tool for tool in iterable_tools},
                    **{tool.name: tool for tool in tools},
                }

                match suggest:
                    case None:
                        suggest_call = False

                    case bool() as suggestion:
                        suggest_call = suggestion

                    case str() as tool_name:
                        suggest_call = tools_mapping[tool_name]

                    case Tool() as tool:
                        assert tool.name in tools_mapping  # nosec: B101
                        suggest_call = tool

                calls_limit = repeated_calls_limit or 3
                metadata = Meta.of(meta)

        return cls(
            tools=tools_mapping,
            suggest_call=suggest_call,
            repeated_calls_limit=calls_limit,
            meta=metadata,
        )

    @classmethod
    async def fetched(
        cls,
        *,
        tools: Collection[str] | None = None,
        tags: MetaTags | None = None,
        repeated_calls_limit: int | None = None,
        suggest: str | bool | None = None,
        meta: Meta | MetaValues | None = None,
        **extra: Any,
    ) -> Self:
        selected_tools: Sequence[Tool] = await Tools.fetch(**extra)
        if tools:
            selected_tools = [tool for tool in selected_tools if tool.name in tools]

        if tags:
            selected_tools = [tool for tool in selected_tools if tool.meta.has_tags(tags)]

        return cls.of(
            *selected_tools,
            suggest=suggest,
            repeated_calls_limit=repeated_calls_limit,
            meta=meta,
        )

    tools: Mapping[str, Tool]
    suggest_call: Tool | bool
    repeated_calls_limit: int
    meta: Meta = META_EMPTY

    def tool_selection(
        self,
        *,
        repetition_level: int = 0,
    ) -> LMMToolSpecification | Literal["auto", "required", "none"]:
        if repetition_level >= self.repeated_calls_limit:
            return "none"  # require no tools if reached the limit

        elif repetition_level != 0:
            return "auto"  # require tools only for the first call, use auto otherwise

        elif self.suggest_call is False:
            return "auto"

        elif self.suggest_call is True:
            return "required"

        elif self.suggest_call.available:  # use suggested tool if able
            return {
                "name": self.suggest_call.name,
                "description": self.suggest_call.description,
                "parameters": self.suggest_call.parameters,
            }

        else:
            return "auto"

    def available_tools(
        self,
        *,
        repetition_level: int = 0,
    ) -> Sequence[LMMToolSpecification]:
        if repetition_level >= self.repeated_calls_limit:
            return ()  # provide no tools if reached the limit

        return tuple(
            {
                "name": tool.name,
                "description": tool.description,
                "parameters": tool.parameters,
            }
            for tool in self.tools.values()
            if tool.available
        )

    async def call_tool(
        self,
        name: str,
        /,
        call_id: str,
        arguments: Mapping[str, Any],
    ) -> MultimodalContent:
        if tool := self.tools.get(name):
            return await tool.tool_call(
                call_id,
                **arguments,
            )

        else:
            raise LMMException(f"Requested tool ({name}) is not defined")

    async def respond_all(
        self,
        requests: LMMToolRequests,
        /,
    ) -> LMMToolResponses:
        return LMMToolResponses(
            responses=await gather(
                *[self.respond(request) for request in requests.requests],
                return_exceptions=False,  # toolbox calls handle errors if able
            )
        )

    async def respond(
        self,
        request: LMMToolRequest,
        /,
    ) -> LMMToolResponse:
        if tool := self.tools.get(request.tool):
            try:
                handling: LMMToolResponseHandling
                match tool.handling:
                    case "auto":
                        handling = "result"

                    case "direct":
                        handling = "direct_result"

                return LMMToolResponse(
                    identifier=request.identifier,
                    tool=request.tool,
                    content=await tool.tool_call(  # pyright: ignore[reportPrivateUsage]
                        request.identifier,
                        **request.arguments,
                    ),
                    handling=handling,
                )

            except LMMToolError as error:  # use formatted error, blow up on other exception
                ctx.log_error(
                    "Tool (%s) returned an error",
                    request.tool,
                    exception=error,
                )
                handling: LMMToolResponseHandling
                match tool.handling:
                    case "auto":
                        handling = "error"

                    case "direct":
                        handling = "direct_result"

                return LMMToolResponse(
                    identifier=request.identifier,
                    tool=request.tool,
                    content=error.content,
                    handling=handling,
                )

        else:
            # log error and provide fallback result to avoid blowing out the execution
            ctx.log_error("Requested tool (%s) is not defined", request.tool)
            return LMMToolResponse(
                identifier=request.identifier,
                tool=request.tool,
                content=MultimodalContent.of("ERROR"),
                handling="error",
            )

    def with_tools(
        self,
        tool: Tool,
        /,
        *tools: Tool,
    ) -> Self:
        return self.__class__.of(
            *(tool, *tools, *self.tools.values()),
            suggest=self.suggest_call,
            repeated_calls_limit=self.repeated_calls_limit,
            meta=self.meta,
        )

    @overload
    def filtered(
        self,
        *,
        tools: Collection[str],
    ) -> Self: ...

    @overload
    def filtered(
        self,
        *,
        tags: MetaTags,
    ) -> Self: ...

    def filtered(
        self,
        *,
        tools: Collection[str] | None = None,
        tags: MetaTags | None = None,
    ) -> Self:
        assert tools is None or tags is None  # nosec: B101
        if tools:
            return self.__class__.of(
                *(tool for tool in self.tools.values() if tool.name in tools),
                suggest=self.suggest_call,
                repeated_calls_limit=self.repeated_calls_limit,
                meta=self.meta,
            )

        elif tags:
            return self.__class__.of(
                *(tool for tool in self.tools.values() if tool.meta.has_tags(tags)),
                suggest=self.suggest_call,
                repeated_calls_limit=self.repeated_calls_limit,
                meta=self.meta,
            )

        else:
            return self
